﻿/* タブブラウザコントロール		*/

module browser;

private import std.stdio;
private import std.string;
private import std.socket;
private import std.conv;
private import std.regexp;
private import std.file;
private import std.path;
private import std.uri;
private import std.utf;
private import std.thread;
private import std.base64;
private import std.stream;
private import std.windows.charset;
private import std.c.windows.windows;
private import std.c.time;
private import std.c.stdlib;

private import dwt.all;
private import dwt.internal.ole.win32.com;
private import dwt.ole.win32.ole;
private import dwt.ole.win32.oleautomation;
private import dwt.ole.win32.olecontrolsite;
private import dwt.internal.ole.win32.OAIDL;
private import dwt.internal.ole.win32.OBJIDL;
private import dwt.internal.ole.win32.OLEIDL;
private import dwt.internal.ole.win32.DOCOBJ;
private import dwt.internal.ole.win32.extras;
private import dwt.internal.ole.win32.MSHTMHST;
private import dwt.internal.ole.win32.ifs;

private import dwt.ole.win32.ole;
private import dwt.ole.win32.oleframe;
 
private import dwt.internal.win32.os;
private import dwt.internal.compatibility;
private import dwt.internal.ole.win32.com;
private import dwt.internal.ole.win32.OAIDL;
private import dwt.internal.ole.win32.OBJIDL;
private import dwt.internal.ole.win32.OLEIDL;
private import dwt.internal.ole.win32.DOCOBJ;
private import dwt.internal.ole.win32.extras;
private import dwt.internal.ole.win32.ifs;
private import dwt.graphics.point;
private import dwt.graphics.rectangle;
private import dwt.widgets.composite;
private import dwt.widgets.event;
private import dwt.widgets.listener;
private import dwt.widgets.menu;
private import dwt.widgets.shell;
private import dwt.ole.win32.variant;
private import dwt.internal.ole.win32.comtypes;


private import pes.checksum;
private import pes.date;
private import pes.path;
private import pes.wstring;
private import pes.conv;
private import pes.variant;
private import pes.wrapper;
private import pes.data.sqlite3;
private import pes.data.inifile;
private import pes.net.http.all;
private import pes.net.shingetsu.all;

private import control;
private import icon;
private import config;
private import schema;
private import threadcache;

///version = OLE_COM;

extern(C) {
	BOOL doOrganizeFavoriteFolderDialog(HWND, CHAR*);
	BOOL doAddToFavoriteFolderDialog(HWND, CHAR *path, UINT path_len,
									 CHAR *title, UINT title_len);
}

extern(Windows)
{
	BOOL SHGetSpecialFolderPathA(
	    HWND hwndOwner,
	    LPTSTR lpszPath,
	    int nFolder,
	    BOOL fCreate
	);
	
	const uint CSIDL_FAVORITES		= 0x0006;        // <user name>\Favorites
	const URLMON_OPTION_USERAGENT  	= 0x10000001;

}
extern(C) 
{
	 HRESULT urlMkGetSessionOption(
	    DWORD dwOption,
	    LPVOID pBuffer,
	    DWORD dwBufferLength,
	    DWORD *pdwBufferLength,
	    DWORD dwReserved
	);
	HRESULT urlMkSetSessionOption(
	    DWORD dwOption,
	    LPVOID pBuffer,
	    DWORD dwBufferLength,
	    DWORD dwReserved
	);
}
alias SHGetSpecialFolderPathA SHGetSpecialFolderPath;

static void initializeSession()
{
	char[] user_agent = new char[MAX_PATH];
	char[] current_user_agent, hoven_user_agent;
	uint ug_len = user_agent.length;
	int per;
	
	urlMkGetSessionOption(URLMON_OPTION_USERAGENT,
							user_agent.ptr, user_agent.length, &ug_len, 0);
	current_user_agent = user_agent[0 .. ug_len];
//	MessageBox.showMessage(current_user_agent);
	
//	per = std.string.find(current_user_agent, getVersionString);
//	if (per == -1) {
		per = std.string.find(current_user_agent, ")");
		if (per != -1) {
			hoven_user_agent = std.string.format("%s; %s)",
									current_user_agent[0 .. per], getVersionString());
		} else {
			hoven_user_agent = std.string.format("%s (%s)",
									current_user_agent, getVersionString());
		}
		
//		urlMkSetSessionOption(URLMON_OPTION_USERAGENT,
//			std.string.toStringz(hoven_user_agent), hoven_user_agent.length, 0);
		urlMkSetSessionOption(URLMON_OPTION_USERAGENT,
			std.string.toStringz("A"), 1, 0);
//	}
}

void addFavorite(Shell shell, char[] path, char[] url, char[] title)
{
	path 	= std.string.toString(std.windows.charset.toMBSz(path));
	title	= std.string.toString(std.windows.charset.toMBSz(title));
	
	if (path.length > MAX_PATH || title.length > MAX_PATH) {
		assert(0);
		return;
	}
	BOOL result;
	char[MAX_PATH + 4] szpath, sztitle;
	
	szpath[] = 0;
	sztitle[] = 0;
	szpath[0 .. path.length] = path;
	sztitle[0 .. title.length] = title;
	result = doAddToFavoriteFolderDialog(shell.handle, szpath.ptr, szpath.length,
												sztitle.ptr, sztitle.length);
	if (result) {
		InternetShortcut shortcut;
		shortcut = new InternetShortcut(std.windows.charset.fromMBSz(sztitle.ptr), url);
		shortcut.writeInternetShortcut(std.windows.charset.fromMBSz(szpath.ptr));
		
		reloadFavorites();
	}
}

void organizeFavorite(Shell shell, char[] path)
{
	BOOL result = doOrganizeFavoriteFolderDialog(shell.handle, std.windows.charset.toMBSz(path));
	
	if (result) {
		reloadFavorites();
	}
}

static char[] getSystemFavoriteDirectory()
{
	char[] path = new char[MAX_PATH + 4];
	
	BOOL result = SHGetSpecialFolderPath(null, path.ptr, CSIDL_FAVORITES, TRUE);
	
	return std.windows.charset.fromMBSz(path.ptr);
}

// Variant
alias pes.variant.Variant PESVariant;
alias dwt.ole.win32.variant.Variant Variant;

// テキストコピー
void copyText(char[] text, Display display = null)
{
	if (display is null) {
		display = Display.getCurrent;
	}
	
	Clipboard clip = new Clipboard(display);
	
	TextTransfer transfer = TextTransfer.getInstance();
	Object[] objs;
	objs ~= new StringObj(text);
	Transfer[] tobjs;
	tobjs ~= transfer;
	clip.setContents(objs, tobjs);
	
	clip.dispose();
	delete clip;
}


// 1行編集ダイアログ
class OneLineEditDialog
{
protected:
	char[]	m_shell_title;
	char[]	m_edit_title;
	char[]	m_edit_text;
	char[]	m_exec_text = "設定";
	int		m_text_width;
	int		m_result;
	Text	m_text;
	
	
	void closeHandler(SelectionEvent e)
	{
		Shell shell = cast(Shell)e.cData;
		shell.close();
	}
	
	void okHandler(SelectionEvent e)
	{
		Shell shell = cast(Shell)e.cData;
		shell.close();
		
		m_result 		= DWT.OK;
		m_edit_text 	= m_text.getText();
	}
	
public:
	this(char[] shell_title, char[] edit_title, int width = 200)
	{
		m_shell_title 	= shell_title;
		m_edit_title	= edit_title;
		m_text_width	= width;
	}
	
	int getLimit()
	{
		return m_text_width;
	}
	
	int run(Point point = null)
	{
		// initialize
		m_result = DWT.CANCEL;
		
		
		// open dialog
		Display	display	= Display.getCurrent();
		Shell shell 	= new Shell(display, DWT.CLOSE|DWT.BORDER|DWT.TITLE|DWT.APPLICATION_MODAL);
		
		shell.setText(m_shell_title);
		shell.setLayout(new GridLayout(2, false));
		
		with (new Label(shell, DWT.NONE)) {
			setText(m_edit_title);
			setLayoutData(new GridData(GridData.BEGINNING));
		}
		with (m_text = new Text(shell, DWT.BORDER|DWT.SINGLE)) {
			setText(m_edit_text);
			setTextLimit(getLimit);
			setLayoutData(new GridData(m_text_width, -1));
		}
		
		Composite page;
		with (page = new Composite(shell, DWT.NONE)) {
			GridData grid = new GridData(GridData.HORIZONTAL_ALIGN_END);
			grid.horizontalSpan = 2;
			setLayoutData(grid);
			
			setLayout(new GridLayout(2, true));
		}
		Button set_button;
		with (set_button = new Button(page, DWT.PUSH)) {
			setText(m_exec_text);
			setLayoutData(new GridData(GridData.END));
			handleSelection(shell, &okHandler);
			// defualt button
			shell.setDefaultButton(set_button);
		}
		
		Button close_button;
		with (close_button = new Button(page, DWT.PUSH)) {
			setText("閉じる");
			setLayoutData(new GridData(GridData.END));
			
			handleSelection(shell, &closeHandler);
		}
		if (point !is null) {
			shell.setLocation(point);
		}
		shell.pack();
		shell.layout();
		shell.open();
		// event loop
		while (! shell.isDisposed()) {
			if (!display.readAndDispatch()) {
				display.sleep();
			}
		}
		
		return m_result;
	}
	
	char[] getText()
	{
		return m_edit_text;
	}
	
	void setText(char[] text)
	{
		m_edit_text = text;
	}
	
	void setButtonText(char[] text)
	{
		m_exec_text = text;
	}
	
	char[] getButtonText()
	{
		return m_exec_text;
	}
}

class NewThreadEditDialog: OneLineEditDialog
{
	this(char[] shell_title, char[] edit_title, int width = 200)
	{
		super(shell_title, edit_title, width);
	}
	
	override int getLimit()
	{
		return 30;
	}
}

alias dwt.internal.ole.win32.OAIDL COM;
alias dwt.internal.ole.win32.comapi COMAPI;

/** ブラウザコントロール */
class CBrowser: Browser
{
protected:
	bool m_initialize = false;
	
	const int onhelp = 0x8001177d;
	const int onclick = 0x80011778;
	const int ondblclick = 0x80011779;
	const int onkeyup = 0x80011776;
	const int onkeydown = 0x80011775;
	const int onkeypress = 0x80011777;
	const int onmouseup = 0x80011773;
	const int onmousedown = 0x80011772;
	const int onmousemove = 0x80011774;
	const int onmouseout = 0x80011771;
	const int onmouseover = 0x80011770;
	const int onreadystatechange = 0x80011789;
	const int onafterupdate = 0x80011786;
	const int onrowexit= 0x80011782;
	const int onrowenter = 0x80011783;
	const int ondragstart = 0x80011793;
	const int onselectstart = 0x80011795;
	
	const int DownloadComplate = 104;
//	EventDispatch myDispatch;
	override void handleEvent(OleEvent e)
	{
/* // イベントブリッヂ処理 (使わないのでコメントアウト)
		if (e.type == DownloadComplate) { // DownloadComplete
			char[][] args;
			args ~= "Document";
			int[] htmlDocumentID = automation.getIDsOfNames(args); 
			if (htmlDocumentID == null) return;
			Variant pVarResult = automation.getProperty(htmlDocumentID[0]);
			if (pVarResult == null || pVarResult.getType() == 0) return;
			//IHTMLDocument2
			OleAutomation htmlDocument = pVarResult.getAutomation();
			pVarResult.dispose();
			
			// Request to be notified of click, double click and key down events
			myDispatch = new EventDispatch(EventDispatch.onkeydown);
			IDispatch idispatch = myDispatch.getAddress;//new IDispatch(myDispatch);
			
			///xxxxxxxxxxxxx
			idispatch.AddRef();
			
			Variant dispatch = new Variant(idispatch);
			htmlDocument.setProperty(EventDispatch.onkeydown, dispatch);
			dispatch.dispose();
			
			//Remember to release OleAutomation Object
			htmlDocument.dispose();
			htmlDocument = null;
		}
*/
		super.handleEvent(e);
	}
	
public:
	void dispose()
	{
/*
		if (myDispatch) {
			myDispatch.dispose();
			myDispatch=null;
		}
*/
		super.dispose();
	}
	this(Composite parnet, int style)
	{
		super(parnet, style);
		
		// site.addEventListener(DownloadComplate, this);
	}
	
	override int setUrl(char[] url)
	{
		return super.setUrl(url);
	}
	
	void execute(char[] script)
	{
		if (script.length == 0) {
			return;
		}
		/* get IHTMLDocument2 */
		char[][] args;
		
		args ~= "Document";
		int[] rgdispid = automation.getIDsOfNames(args); //$NON-NLS-1$
		int dispIdMember = rgdispid[0];
		Variant pVarResult = automation.getProperty(dispIdMember);
		if (pVarResult == null || pVarResult.getType() == COM.VT_EMPTY) return false;
		OleAutomation document = pVarResult.getAutomation();
		pVarResult.dispose();
		
		/* get IHTMLWindow2 */
		args = null;
		args ~= "parentWindow";
		rgdispid = document.getIDsOfNames(args); //$NON-NLS-1$ 
		dispIdMember = rgdispid[0];
		pVarResult = document.getProperty(dispIdMember);
		OleAutomation ihtmlWindow2 = pVarResult.getAutomation();
		pVarResult.dispose();
		document.dispose();
		
		args = null;
		args ~= "execScript";
		args ~= "code";
		rgdispid = ihtmlWindow2.getIDsOfNames(args); //$NON-NLS-1$  //$NON-NLS-2$
		Variant[] rgvarg = new Variant[1];
		rgvarg[0] = new Variant(script);
		int[] rgdispidNamedArgs = new int[1];
		rgdispidNamedArgs[0] = rgdispid[1];
		pVarResult = ihtmlWindow2.invoke(rgdispid[0], rgvarg, rgdispidNamedArgs);
		rgvarg[0].dispose();
		ihtmlWindow2.dispose();
		if (pVarResult == null) return false;
		pVarResult.dispose();
		
		return true;
	}
	
	void organizeFavorites()
	{
		const char[] script = 	`var shellUIHelper = new ActiveXObject("Shell.UIHelper");`
								`shellUIHelper.showBrowserUI("OrganizeFavorites", null);`;
		
		execute(script);
	}
	
	void addFavorite()
	{
		const char[] script = 	`var shellUIHelper = new ActiveXObject("Shell.UIHelper");`
								`shellUIHelper.AddFavorite(document.URL, document.title);`;
		
		execute(script);
	}
	
	void addFavorite(char[] url, char[] title)
	{
		const char[] fmt = 		`var shellUIHelper = new ActiveXObject("Shell.UIHelper");`
								`shellUIHelper.AddFavorite('%s','%s');`;
		
		execute(std.string.format(fmt, escapeString(url), escapeString(title)));
	}
	
	static char[] escapeString(char[] unescaped_string)
	{
		return std.string.replace(unescaped_string, "'", "\\'");
	}
	
/++ イベントブリッジ 
	// EventDispatch implements a simple IDispatch interface which will be called 
	// when the event is fired.
	class EventDispatch {
		private IDispatch iDispatch;
		private int refCount = 0;
		private int eventID;
		
		const int onhelp = 0x8001177d;
		const int onclick = 0x80011778;
		const int ondblclick = 0x80011779;
		const int onkeyup = 0x80011776;
		const int onkeydown = 0x80011775;
		const int onkeypress = 0x80011777;
		const int onmouseup = 0x80011773;
		const int onmousedown = 0x80011772;
		const int onmousemove = 0x80011774;
		const int onmouseout = 0x80011771;
		const int onmouseover = 0x80011770;
		const int onreadystatechange = 0x80011789;
		const int onafterupdate = 0x80011786;
		const int onrowexit= 0x80011782;
		const int onrowenter = 0x80011783;
		const int ondragstart = 0x80011793;
		const int onselectstart = 0x80011795;

		this(int eventID) {
			this.eventID = eventID;
			createCOMInterfaces();
		}
		IDispatch getAddress() {
			return iDispatch;
		}
		private void createCOMInterfaces() {
			int[] arg00 = new int[7];
			static int[7] arg00data = [2, 0, 0, 1, 3, 4, 8];
			
			arg00[0 .. length] = arg00data;
			
			class Local00: IDispatch
			{
				private import std.c.windows.windows;
				private import std.c.windows.com;
				int refCount = 0;
				EventDispatch parent;
				
				this(EventDispatch parent) { this.parent = parent;}
				
			extern (Windows) :
				HRESULT QueryInterface(REFIID riid, void **ppvObject){
					return parent.QueryInterface(riid, ppvObject);
				}
				ULONG AddRef()	{ return parent.AddRef(); }
				ULONG Release()	 { return parent.Release();}
				HRESULT GetTypeInfoCount(UINT * pctinfo) { return COM.E_NOTIMPL; }
				HRESULT GetTypeInfo(UINT iTInfo, LCID lcid, ITypeInfo * ppTInfo) { return COM.E_NOTIMPL; }
				HRESULT GetIDsOfNames(REFIID riid, LPOLESTR * rgszNames, UINT cNames, LCID lcid, DISPID * rgDispId) 
				{
					return COM.E_NOTIMPL; 
				}
				HRESULT Invoke(DISPID dispIdMember,REFIID riid,LCID lcid,
								WORD wFlags,DISPPARAMS* pDispParams,
								VARIANT* pVarResult,EXCEPINFO* pExcepInfo,UINT* puArgErr)
				{
					return parent.Invoke(dispIdMember, riid, lcid, wFlags, pDispParams, pVarResult, pExcepInfo, puArgErr);
				}
			}
			
			iDispatch = cast(IDispatch)new Local00(this);
		}
		private void disposeCOMInterfaces() {
			if (iDispatch !is null) {
				iDispatch.Release();
			}
			iDispatch = null;
		}
		private int AddRef() {
			refCount++;
			return refCount;
		}
		private HRESULT Invoke(DISPID dispIdMember,REFIID riid,LCID lcid,
								WORD wFlags,DISPPARAMS* pDispParams,
								VARIANT* pVarResult,EXCEPINFO* pExcepInfo,UINT* puArgErr)
		{
			switch(eventID) {
/*
			case onhelp: MessageBox.showMessage("onhelp"); break;
			case onclick: MessageBox.showMessage("onclick"); break;
			case ondblclick: MessageBox.showMessage("ondblclick"); break;
			case onkeyup: MessageBox.showMessage("onkeyup"); break;
			case onkeydown: MessageBox.showMessage("onkeydown"); break;
			case onkeypress: MessageBox.showMessage("onkeypress"); break;
			case onmouseup: MessageBox.showMessage("onmouseup"); break;
			case onmousedown: MessageBox.showMessage("onmousedown"); break;
			case onmousemove: MessageBox.showMessage("onmousemove"); break;
			case onmouseout: MessageBox.showMessage("onmouseout"); break;
			case onmouseover: MessageBox.showMessage("onmouseover"); break;
			case onreadystatechange: MessageBox.showMessage("onreadystatechange"); break;
			case onafterupdate: MessageBox.showMessage("onafterupdate"); break;
			case onrowexit: MessageBox.showMessage("onrowexit"); break;
			case onrowenter: MessageBox.showMessage("onrowenter"); break;
			case ondragstart: MessageBox.showMessage("ondragstart"); break;
			case onselectstart: MessageBox.showMessage("onselectstart"); break;
*/
			default:
				break;
			}
			return COM.S_OK;
		}
		private int QueryInterface(REFIID riid, void **ppvObject) {
			if (riid == null || ppvObject == null) return COM.E_INVALIDARG;
			GUID *guid = new GUID;
//			COM.MoveMemory(guid, cast(void*)riid, GUID.sizeof);
			*guid = *cast(GUID*)riid;
			if (COMAPI.IsEqualGUID(guid, &dwt.internal.ole.win32.com.COM.IIDIUnknown) 
				|| COMAPI.IsEqualGUID(guid, &dwt.internal.ole.win32.com.COM.IIDIDispatch)) 
			{
				*ppvObject = iDispatch;
				AddRef();
				return COM.S_OK;
			}
//			COM.MoveMemory(ppvObject, arg00, 4);
			*ppvObject = null;
			return COM.E_NOINTERFACE;
		}
		int Release() {
			refCount--;
			if (refCount == 0) {
				disposeCOMInterfaces();
			}
			return refCount;
		}
		
		void dispose()
		{
			disposeCOMInterfaces();
		}
	}
+/
}

//alias Browser CBrowser;

// ブラウザタブベース
abstract class BrowserTabPage
{
private:
	bool							m_disposed;
	void delegate()					m_status_changed;
	void delegate(BrowserTabPage) 	m_set_parameter;
	
protected:
	CTabFolder	m_tab_folder;
	CTabItem 	m_item;
	int			m_style;
	
	Composite 	m_page;
	
	Label		m_status_bar;
	CCombo		m_address_bar;
	
	void setTabItem(CTabItem item)
	{
		m_item = item;
	}
	
	void setTabFolder(CTabFolder folder)
	{
		m_tab_folder = folder;
	}
	
	void setStyle(int style)
	{
		m_style = style;
	}
	
	int getStyle()
	{
		return m_style;
	}
	
	void setPage(Composite page)
	{
		m_page = page;
	}
	
	abstract void createWidgets();
	
	void disposeHandler(DisposeEvent e)
	{
		dispose();
	}
	
public:
	this(char[] address, CTabFolder folder, int style)
	{
		CTabItem 	item = new CTabItem(folder, DWT.NONE);
		Composite 	page = new Composite(folder, style);
		
		setTabFolder(folder);
		setTabItem(item);
		setStyle(style);
		setPage(page);
		createWidgets();
		
		m_disposed = false;
		item.handleDispose(null, &disposeHandler);
		item.setControl(getPage);
		
		folder.setSelection(item);
	}
	
	bool isSelective()
	{
		return getTabFolder().getSelection() is getTabItem();
	}
	
	void setStatusBar(Label status_bar)
	{
		m_status_bar = status_bar;
	}
	
	void setStatusBarText(char[] status)
	{
		if (m_status_bar is null) {
			return;
		}
		if (m_tab_folder.getSelection is m_item) {
			m_status_bar.setText(status);
		}
	}
	
	void setAddressBar(CCombo address_bar)
	{
		m_address_bar = address_bar;
	}
	
	int compareAddress(char[] address)
	{
		return std.string.cmp(getAddress(), address);
	}
	
	void setAddressBarText(char[] address)
	{
		if (m_address_bar is null) {
			return;
		}
		if (m_tab_folder.getSelection is m_item) {
			m_address_bar.setText(address);
		}
	}
	
	void handleStatusChanged(void delegate() func)
	{
		m_status_changed = func;
	}
	
	void changeStatus()
	{
		if (m_status_changed != null) {
			m_status_changed();
		}
	}
	
	void handleSetParameter(void delegate(BrowserTabPage) func)
	{
		m_set_parameter = func;
	}
	
	void setPageParameter(BrowserTabPage page)
	{
		if (m_set_parameter != null) {
			m_set_parameter(page);
		}
	}
	
	Composite getPage()
	{
		return m_page;
	}
	
	Composite getControl()
	{
		return m_page;
	}
	
	abstract TypeInfo getType();
	abstract void setAddress(char[] address);
	abstract char[] getAddress();
	abstract void refresh();
	abstract void forward();
	abstract void back();
	abstract void stop();
	abstract bool isStopEnabled();
	abstract bool isForwardEnabled();
	abstract bool isBackEnabled();
	
	BrowserTabPage getCurrentTabPage()
	{
		CTabItem item = getTabFolder.getSelection;
		
		if (item !is null) {
			return cast(BrowserTabPage)item.getData();
		}
		
		return null;
	}
	
	CBrowser getBrowser()
	{
		return null;
	}
	
	void dispose()
	{
		if (!m_disposed) {
			m_tab_folder	= null;
			m_item			= null;
			m_page			= null;
			m_status_bar	= null;
			m_address_bar	= null;
			m_disposed		= true;
		}
	}
	
	void setTitleImage(Image image)
	{
		Image current = getTabItem.getImage();
		
		if (current is image) {
			return;
		}
		
		CTabFolder tab = getTabFolder;
		tab.setRedraw(false);
		scope(exit) tab.setRedraw(true);
		
		getTabItem.setImage(image);
	}
	
	void setTitleText(char[] text)
	{
		char[] current = getTabItem.getText();
		
		if (std.string.cmp(current, text) == 0) {
			return;
		}
		
		CTabFolder tab = getTabFolder;
		tab.setRedraw(false);
		scope(exit) tab.setRedraw(true);
		
		getTabItem.setText(text);
	}
	
	char[] getTitleText()
	{
		return getTabItem.getText();
	}
	
	CTabFolder getTabFolder()
	{
		return m_tab_folder;
	}
	
	CTabItem getTabItem()
	{
		return m_item;
	}
	
	void setData(char[] key, Object data)
	{
		getPage.setData(key, data);
	}
	
	Object getData(char[] key)
	{
		return getPage.getData(key);
	}
	
	BrowserTabPage addPage(char[] address, int style = DWT.NONE)
	{
		BrowserTabPage tab = createTabPage(address, m_tab_folder, style);
		setPageParameter(tab);
		
		return tab;
	}
}

BrowserTabPage createTabPage(char[] address, CTabFolder folder, int style)
{
	if (std.string.find(address, "shingetsu://") == 0) {
		// shingetsu browser
		return new ShingetsuBrowserPage(address, folder, style);
	}
	
	// web browser
	return new WebBrowserTabPage(address, folder, style);
}

// Loader Process Controler
class ProcessControler: Runnable {
	uint		count;
	PageLoader 	loader;
	this(PageLoader loader) { this.loader = loader; }
	void run()
	{
		if (loader.isDisposed) {
			return;
		}
		if (!loader.isBusy) {
			char[] message = loader.errorMessage();
			if (message.length != 0) {
				// エラー
				loader.drawErrorPage(message);
			} else {
				loader.drawPage();
			}
		} else {
//			if (++count > 600) {
//			// エラー
//				loader.setTitleImage(getIcon("error"));
//				loader.drawErrorPage("TimeOut");
//			}
			if ((count++ % 10) == 0) {
				loader.progressChanged();
			}
			Display.getCurrent.timerExec(50, this);
		}
	}
}

class PageLoader
{
protected:
	char[]	m_address;
	char[]	m_error_message;
	
	Schema m_schema;
	ShingetsuBrowser m_parent;
	Thread			m_thread;
	ThreadMessage 	m_data;
	ServerIndex[] 	m_internal_indexes;
	ThreadMessage[]	m_messages;
	char[]			m_recent_id;
	
	class ServerIndex {
		time_t			time;
		char[]			md5;
	}
	
	void setAttach(ThreadMessage message, char[] attachment)
	{
		ubyte[] bin		= cast(ubyte[])std.base64.decode(attachment);
		scope(exit) {
			delete bin;
		}
		message.setAttachName(bin);
		
		char[] 	path 	= message.getAttachPath();
		File	file	= new File(path, FileMode.OutNew);
		
		file.writeBlock(bin.ptr, bin.length);
		file.close();
	}
	
	void headerResponseHandler(ThreadDataEvent event)
	{
		ServerIndex index = new ServerIndex;
		
		index.time 	= event.time;
		index.md5	= event.md5.dup;
		
		m_internal_indexes ~= index;
	}
	
	void messageResponseHandler(ThreadDataEvent event)
	{
		Schema schema = cast(Schema)event.cData;
		ThreadMessage message = new ThreadMessage;
		
		message.title		= schema.threadName;
		message.time 		= event.time;
		message.md5			= event.md5.dup;
		
		if ("body" in event.elements) {
			message.message = event.elements["body"].dup;
		}
		
		if ("remove_id" in event.elements) {
			message.removeId = event.elements["remove_id"];
		}
		if ("remove_stamp" in event.elements) {
			message.removeStamp = event.elements["remove_stamp"];
		}
		
		if ("mail" in event.elements) {
			message.mail = event.elements["mail"].dup;
		}
		if ("name" in event.elements) {
			message.name = event.elements["name"].dup;
		}
		if ("attach" in event.elements) {
			if ("suffix" in event.elements) {
				message.ext = event.elements["suffix"].dup;
				// 保存
				setAttach(message, event.elements["attach"]);
			} else {
				message.ext = null;
			}
		}
		if ("pubkey" in event.elements) {
			char[] short_key;
			
			message.pubkey = event.elements["pubkey"].dup;
			try {
			short_key = message.pubkey;
			if (short_key.length > 512) {
				short_key = short_key[0 .. 512];
			}
			message.trip = std.base64.encode(cast(char[])md5Sum(short_key))[0 .. 11];
			} catch (Exception e) {
				writef("%s\n", e.toString);
			}
		}
		if ("sign" in event.elements) {
			message.sign = event.elements["sign"].dup;
		}
		
		m_messages ~= message;
	}
	
	void getItemFromNetwork(Schema schema)
	{
		ThreadCache		cache			= ThreadCache.getInstance();
		ServerIO		server_io		= new ServerIO(new InternetAddress(getConfig.nodeName, getConfig.nodePort));
		time_t			base			= 0;
		time_t			last_access		= 0;
		ThreadIndex		index 			= cache.getIndex(schema.threadName);
		ServerIndex[] 	new_indexes;
		
		m_internal_indexes = null;
		m_messages = null;
		
		if (index is null 
			|| (index.flag & 0x01) == 0
			|| !server_io.haveThread(schema.threadName)) 
		{
			// 新規
			ThreadIO io = new ThreadIO(new InternetAddress(getConfig.nodeName, getConfig.nodePort));
			io.registerThread(schema.threadName);
			delete io;
		} else {
			m_recent_id = index.id().dup;
			last_access = index.lastAccessTime;
			base = index.lastPostTime - 3600 * 2;	// 2時間
//			base = 0;
		}
		
		ThreadMessageSummary[] summaries = cache.getRecentMessageSummaries(schema.threadName, base);
		server_io.getThreadHeader(null, &headerResponseHandler, schema.threadName, base);
		
		bool found;
		foreach (i; m_internal_indexes) {
			found = false;
			
			foreach (j; summaries) {
				if (i.time == j.time
					&& std.string.icmp(i.md5, j.md5) == 0)
				{
					found = true;
					break;
				}
			}
			if (!found && i.time < last_access) {
				new_indexes ~= i;
			}
		}
		// last_access以降を取得
		server_io.getThreadData(schema, &messageResponseHandler, schema.threadName, last_access);
		
		// last_access以前(遅)を取得
		foreach (e; new_indexes) {
			server_io.getThreadData(schema, &messageResponseHandler, schema.threadName, e.time, e.md5);
		}
		
		cache.insertMessages(m_messages);
		/*
		foreach (inout e; summaries) {
			e.dispose();
			e = null;
		}
		delete summaries;
		*/
		if (m_messages.length != 0) {
			foreach (inout e; m_messages) {
				e.dispose();
			}
			delete m_messages;
		}
		m_messages = null;
	}
	
	ThreadMessage[] loadData(Schema schema, out uint begin, out uint count)
	{
		ThreadCache cache = ThreadCache.getInstance();
		ThreadMessage[] messages;
		
		count = cache.getMessageCount(schema.threadName);
		
		if (schema.hash.length != 0) {
			ThreadMessage message = cache.getMessage(schema.threadName, schema.hash);
			if (message !is null) {
				messages ~= message;
			}
		} else {
			switch (schema.mode) {
			case Schema.Mode.BACK_ROWS:
				if (count < schema.items) {
					begin = 0;
				} else {
					begin = count - schema.items;
				}
				messages = cache.getMessages(schema.threadName, begin, schema.items);
				break;
			case Schema.Mode.PAGES:
				begin 	= schema.items * schema.page;
				messages = cache.getMessages(schema.threadName, begin, schema.items);
				break;
			default:
				begin = 0;
				messages = cache.getMessages(schema.threadName, begin);
				break;
			}
		}
		
		return messages;
	}
	
public:
	this(ShingetsuBrowser parent, char[] address)
	{
		m_address 	= address;
		m_parent	= parent;
	}
	
	bool isDisposed()
	{
		return m_parent.isDisposed ? true:false;
	}
	
	void setParentParameter()
	{
		m_schema = Schema.parse(m_address);
		
		m_parent.setTitleText(m_schema.threadName);
		m_parent.setAddressBarText(m_address);
		m_parent.progressChanged();
//		m_parent.drawLoadingPage(null);
	}
	
	void fromNetwork()
	{
		setParentParameter();
		
//		debug {
//			loadFromNetwork();
//		} else {
			m_thread = new Thread(&loadFromNetwork);
			m_thread.start();
//		}
		// set timer
		Display.getCurrent.timerExec(50, new ProcessControler(this));
	}
	
	void fromCache()
	{
		setParentParameter();
		drawPage();
	}
	
	bool isBusy()
	{
//		debug {
//			return false;
//		} else {
			return m_thread.getState != Thread.TS.TERMINATED;
//		}
	}
	
	void wait()
	{
		m_thread.wait();
	}
	
	int loadFromNetwork()
	{
		try {
			getItemFromNetwork(m_schema);
		} catch (Exception e) {
			m_error_message = e.toString;
		}
		
		return 0;
	}
	
	void drawPage()
	{
		uint count, begin_no;
		
		setTitleImage(getIcon("saku"));
		
		ThreadMessage[] messages = loadData(m_schema, begin_no, count);
		char[] page = ThreadMessage.toXML(m_schema, getConfig.themeDirectory, messages, begin_no, m_recent_id);
		
		m_parent.drawPage(page,  count);
		delete page;
		foreach (e; messages) {
			e.dispose();
		}
		delete messages;
	}
	
	void drawLoadingPage(char[] name)
	{
		m_parent.drawLoadingPage(name);
	}
	
	void drawErrorPage(char[] message)
	{
		setTitleImage(getIcon("error"));
		m_parent.drawErrorPage(message);
	}
	
	char[] errorMessage()
	{
		return m_error_message;
	}
	
	void progressChanged()
	{
		m_parent.progressChanged();
	}
	
	void setTitleImage(Image image)
	{
		m_parent.setTitleImage(image);
	}
}

class ShingetsuBrowser: CBrowser
{
protected:
	const char[] LOADING_PAGE =
"<html><head><title>Loading</title></head><body><h1>読み込み中でございます..</h1></body></html>";
	
	const char[] ERROR_PAGE =
"<html><head><title>Error</title></head><body><h1>エラーです!: %s</h1></body></html>";

	const int	TEMPORARY_BUFFER = 100;
	
	int						m_current_index;
	char[][int] 			m_history;
	Schema					m_current_schema;
	ShingetsuBrowserPage 	m_page;
	uint 					m_unique;
	uint					m_current_count;
	uint					m_clear_counter;
	
	void cleanTemporaryFiles()
	{
		if (m_clear_counter++ % TEMPORARY_BUFFER) {
			return;
		}
		char[] tempdir = getConfig.temporaryDirectory;
		char[][] samples = std.file.listdir(tempdir, "*.xml");
		
		scope(exit)
		{
			foreach (inout s; samples) {
				delete s;
				s = null;
			}
			delete samples;
		}
		
		uint count = samples.length;
		foreach (xml; samples.sort) { // 古いほうから
			if (count-- < TEMPORARY_BUFFER) {
				break; // 100ずつ
			}
			try {
				std.file.remove(xml);
			} catch (FileException) {
				// do nothing
			}
		}
	}
	
	class LocationChangeListener: LocationListener
	{
		void changing(LocationEvent e)
		{
			if (std.string.find(e.location, "shingetsu://") == 0) {
				// current window
//				e.doit = false;
//				setUrl(e.location); 
				// ひとつ上で処理
			}
		}
		
		void changed(LocationEvent e)
		{
			
		}
	}
	
public:
	this(Composite parent, int style, ShingetsuBrowserPage page = null)
	{
		super(parent, style);
		m_page = page;
		m_unique = cast(uint)time(null);
		
		addLocationListener(new LocationChangeListener);
	}
	
	override int setUrl(char[] address)
	{
		if (std.string.find(address, "shingetsu://") == 0) {
			m_history[++m_current_index] = address;
			m_current_schema	= Schema.parse(address);
			
			setAddressBarText(m_history[m_current_index]);
			
			try {
				PageLoader loader = new PageLoader(this, address);
				if (std.string.icmp(m_current_schema.method, "get") == 0) {
					// getの場合のみ、ネットワークから取得
					loader.fromNetwork(); // thread
				} else {
					// キャッシュから取得
					loader.fromCache();
				}
			} catch (Exception e) {
				drawErrorPage(e.toString);
			}
		} else {
			super.setUrl(address);
		}
		
		return 0;
	}
	
	override char[] getUrl()
	{
		if (m_current_index in m_history) {
			return m_history[m_current_index];
		}
		return super.getUrl();
	}
	
	override int forward()
	{
		++m_current_index;
		super.forward();
		setAddressBarText(getUrl);
	}
	
	override int back()
	{
		if (m_current_index != 1) {
			--m_current_index;
			super.back();
			setAddressBarText(getUrl);
		}
	}
	
	void drawPage(char[] page, uint count)
	{
		m_current_count = count;
		
		Schema schema = Schema.parse(getUrl);
		char[] jump;
		
		if (schema.top.length != 0) {
			jump = schema.top;
		}
		
		cleanTemporaryFiles();
		
		char[] path = std.path.join(getConfig.temporaryDirectory,
					std.string.format("%s%s%s.xml", time(null),
										(rand() % 0x1000000) >> 8, m_unique));
		File file = new File(path, FileMode.OutNew);
		
		file.writeBlock(page.ptr, page.length);
		file.close();
		
		if (jump.length != 0) {
			path ~= "#" ~ jump;
		}
		
		setUrl(path);
	}
	
	uint getCurrentItemCount()
	{
		return m_current_count;
	}
	
	void drawErrorPage(char[] message)
	{
		char[] error_page = std.string.format(ERROR_PAGE, message);
		setText(error_page);
	}
	
	void drawLoadingPage(char[] name)
	{
		setText(LOADING_PAGE);
	}
	
	void progressChanged(uint dummy = 0)
	{
		if (m_page !is null) {
			m_page.progressChanged();
		}
	}
	
	void setTitleText(char[] title)
	{
		if (m_page !is null) {
			m_page.setTitleText(title);
			
			if (m_current_schema.hash.length != 0) {
				m_page.setTitleDetail(m_current_schema.hash);
			} else {
				switch (m_current_schema.mode) {
				case Schema.Mode.UNDEF:
					break;
				case Schema.Mode.PAGES:
					m_page.setTitleDetail(std.string.format("%sページ", m_current_schema.page + 1));
					break;
				case Schema.Mode.BACK_ROWS:
					m_page.setTitleDetail(std.string.format("最新%s件", m_current_schema.items));
					break;
				default:
					break;
				}
			}
		}
	}
	
	void setAddressBarText(char[] address)
	{
		if (m_page !is null) {
			m_page.setAddressBarText(address);
		}
	}
	
	void setTitleImage(Image img)
	{
		if (m_page !is null) {
			m_page.setTitleImage(img);
		}
	}
}


class BorwserChildWindow: ControlBase
{
protected:
	static uint		s_child_count;
	static Object	s_lock;
	
	static this()
	{
		s_lock 			= new Object;
		s_child_count	= 0;
	}
	
protected:
	Point			m_location;
	Point 			m_size;
	Shell 			m_shell;
	Display			m_display;
	CBrowser 		m_browser;
	BrowserTabPage	m_parent;
	Runnable		m_finalizer;
	
	bool			m_onetime;
	
	class TitleChangeListener: TitleListener
	{
		void changed(TitleEvent e)
		{
			m_shell.setText(e.title);
		}
	}
	
	class CloseListener: CloseWindowListener
	{
		void close(WindowEvent e)
		{
			m_shell.close();
		}
	}
	
	class LocationChangeListener: LocationListener
	{
		uint count = 0;
		void changing(LocationEvent e)
		{
			if (e.doit == false) {
				return;
			}
			
			if (m_onetime && count++ > 0) {
				close();
			}
			
			if (std.string.ifind(e.location, "shingetsu:") == 0) {
				e.doit = false;
				m_browser.setUrl(e.location);
				return;
			}
			
			if (std.string.ifind(e.location, "file:///") != 0
				&& std.string.ifind(e.location, "javascript:") != 0
				&& std.string.ifind(e.location, getConfig.temporaryDirectory) == -1) 
			{
				if (m_parent !is null) {
					BrowserTabPage tab = m_parent.addPage(e.location);
					tab.setAddress(e.location);
					e.doit = false;
				}
			}
		}
		
		void changed(LocationEvent e)
		{
		}
	}
	
	class NewWindowListener: OpenWindowListener
	{
		void open(WindowEvent e)
		{
			BorwserChildWindow child = new BorwserChildWindow;
			child.setParent(m_parent);
			e.browser	= child.getBrowser();
			child.run();
		}
	}
	
	void closeHandler(SelectionEvent e)
	{
		m_shell.close();
	}
	
	void shellCloseHandler(ShellEvent e)
	{
		m_location	= m_shell.getLocation;
		m_size		= m_shell.getSize;
	}
	
	void disposeHandler(DisposeEvent e)
	{
		if (m_finalizer !is null) {
			m_finalizer.run();
		}
		if (m_browser !is null) {
			m_browser.dispose();
			m_browser = null;
		}
		finalize();
	}
	
	void finalize()
	{
		synchronized(s_lock) {
			if (--s_child_count == 0) {
				serialize();
			}
		}
	}
	
	void serialize()
	{
		setGUIConfig(
			VArray.make(
			cast(long)m_size.x,
			cast(long)m_size.y,
			cast(long)m_location.x,
			cast(long)m_location.y)
		);
	}
	
	void deserialize()
	{
		long lx, ly, sx, sy;
		
		try {
			VArray array = getGUIConfig();
			if (array.toLong == 4) {
				sx = array[0].toLong;
				sy = array[1].toLong;
				lx = array[2].toLong;
				ly = array[3].toLong;
			}
// parent x
//			synchronized(s_lock) {
				uint startx = lx + ((s_child_count / 5) * 10);
				uint starty = ly + ((s_child_count / 5) * 10);
				
				lx = startx + (s_child_count % 5) * 20;
				ly = starty + (s_child_count % 5) * 20;
//			}
			m_shell.setSize(sx, sy);
			m_shell.setLocation(lx, ly);
			
		}  catch (SQLiteException e) {
			m_shell.setSize(400, 300);
			return; // first
		}
	}
	
	Image getShellIcon()
	{
		Image icon = getIcon("comment");
		icon.transparentPixel = 0xffffff;
		
		return icon;
	}
	
public:
	this()
	{
		m_display	= Display.getDefault();
		m_shell 	= new Shell(m_display, 
									DWT.CLOSE|DWT.MIN|DWT.BORDER|DWT.TITLE|DWT.PRIMARY_MODAL
									| DWT.RESIZE|DWT.ON_TOP);
//		m_shell.setImage(getShellIcon());
		m_shell.setLayout(new FillLayout);
		m_shell.handleShellClose(null, &shellCloseHandler);
		
		Composite panel = new Composite(m_shell, DWT.NONE);
		panel.setLayout(new GridLayout(1, false));
		panel.handleDispose(null, &disposeHandler);
		
		with (m_browser = new ShingetsuBrowser(panel, DWT.FLAT)) {
			addTitleListener(new TitleChangeListener);
			addCloseWindowListener(new CloseListener);
			addLocationListener(new LocationChangeListener);
			addOpenWindowListener(new NewWindowListener);
			setLayoutData(new GridData(GridData.FILL_BOTH));
		}
		
		Composite page;
		with (page = new Composite(panel, DWT.NONE)) {
			setLayoutData(new GridData(GridData.HORIZONTAL_ALIGN_END));
			setLayout(new GridLayout(1, true));
		}
		Button close_button;
		with (close_button = new Button(page, DWT.PUSH)) {
			setText("閉じる");
			setLayoutData(new GridData(GridData.END));
			
			handleSelection(m_shell, &closeHandler);
			// defualt button
			m_shell.setDefaultButton(close_button);
		}
	}
	
	void run()
	{
		synchronized(s_lock) {
			deserialize();
			++s_child_count;
		}
		if (m_location !is null) {
			m_shell.setLocation(m_location);
		}
		if (m_size !is null) {
			m_shell.setSize(m_size);
		}
		
		m_shell.layout();
		m_shell.open();
	}
	
	void close()
	{
		m_shell.close();
	}
	
	void setFinalizer(Runnable finalizer)
	{
		m_finalizer = finalizer;
	}
	
	void setParent(BrowserTabPage tab)
	{
		m_parent = tab;
	}
	
	void setSize(Point p)
	{
		m_size = p;
	}
	
	void setLocation(Point p)
	{
		m_location = p;
	}
	
	void setAddress(char[] address)
	{
		m_browser.setUrl(address);
	}
	
	char[] getAddress()
	{
		return m_browser.getUrl();
	}
	
	CBrowser getBrowser()
	{
		return cast(CBrowser)m_browser;
	}
	
	void setOneTimeEnabled(bool sw)
	{
		m_onetime = sw;
	}
	
	bool setOneTimeEnabled()
	{
		return m_onetime;
	}
	
}

class BorwserPostWindow: BorwserChildWindow
{
protected:
	uint count;
	class LocationChangeListener2: LocationListener
	{
		void changing(LocationEvent e)
		{
			char[] url = e.location;
			if (count++ > 1) {
				e.doit = false;
				close();
			}
		}
		
		void changed(LocationEvent e)
		{
		}
	}
	
public:
	this()
	{
		super();
		m_browser.addLocationListener(new LocationChangeListener2);
	}
	
	Image getShellIcon()
	{
		return getIcon("pencil_go");
	}
}

class ShingetsuBrowserPage: BrowserTabPage
{
	Composite			m_parent;
	char[]				m_current_address;
	uint				m_sw;
	uint				m_unique;
	CLabel				m_title;
	char[]				m_title_text_origin;
	ShingetsuBrowser	m_browser;
	char[]				m_target;
	PostFinalizer		m_finalizer;
	ToolItem			m_recent_item;
	
	static Menu			s_res_menu;
	
	void clearPage()
	{
		if (m_browser !is null) {
			m_browser.setText("");
		}
	}
	
	// browser event
	class LocationChangeListener: LocationListener
	{
	public:
		void changing(LocationEvent e)
		{
			if (e.doit == false) {
				return;	// ignore
			}
			
			char[] url = e.location;
			
			if (std.string.ifind(e.location, "file:///") != 0
				&& std.string.ifind(e.location, "javascript:") != 0
				&& std.string.cmp(e.location, CBrowser.ABOUT_BLANK) != 0
				&& std.string.ifind(e.location, getConfig.temporaryDirectory) == -1) 
			{
				// タブウインドウ
				BrowserTabPage browser = createTabPage(e.location, getTabFolder, getStyle);
				setPageParameter(browser);
				browser.setAddress(e.location);
				
				e.doit = false;
			}
			// else current
			char[][] parts = std.string.split(e.location, "#");
			if (parts.length == 2) {
				if (parts[1].length == 8) {
					m_target = parts[1];
				}
			}
			if (isSelective) {
				changeStatus();
			}
		}
		
		void changed(LocationEvent e)
		{
			if (isSelective) {
				changeStatus();
			}
		}
	}
	
	class NewWindowListener: OpenWindowListener
	{
		BrowserTabPage parent;
		this(BrowserTabPage parent) { this.parent = parent; }
		public void open(WindowEvent e)
		{
			BorwserChildWindow child = new BorwserChildWindow;
			child.setParent(parent);
			e.browser	= child.getBrowser();
			child.run();
		}
	}
	
	class StatusChangeListener: StatusTextListener
	{
		public void changed(StatusTextEvent e)
		{
			char[] text = e.text;
			
			if (std.string.ifind(text, "file:///") == 0) {
				char[][] id = std.string.split(text, "#");
				if (id.length > 1) {
					setStatusBarText(std.string.format("%s#%s", getTitleText(), id[1]));
					return;
				}
			}
			setStatusBarText(e.text);
		}
	}
	
	void recentItemHandler(SelectionEvent e)
	{
		Schema schema = Schema.parse(getAddress());
		
		schema.page 	= 0;
		schema.mode 	= Schema.Mode.BACK_ROWS;
		schema.items 	= getConfig.items;
		schema.hash		= null;
		schema.method 	= "get";
		
		setAddress(schema.toURI);
	}
	
	void allItemHandler(SelectionEvent e)
	{
		Schema schema = Schema.parse(getAddress());
		
		schema.page 	= 0;
		schema.items 	= 0;
		schema.mode 	= Schema.Mode.UNDEF;
		schema.hash		= null;
		schema.method 	= "get";
		
		setAddress(schema.toURI);
	}
	
	void nextItemHandler(SelectionEvent e)
	{
		Schema schema = Schema.parse(getAddress());
		uint count = m_browser.getCurrentItemCount;
		
		if (schema.mode == Schema.Mode.PAGES) {
			if ((schema.page + 1) * schema.items < count) {
				schema.page = schema.page + 1;
				schema.method 	= "cache";
			} else {
				schema.method 	= "get";
			}
			setAddress(schema.toURI);
		} else {
			uint page		= (count / schema.items);
			if (page > 0) {
				--page;
			}
			
			schema.mode 	= Schema.Mode.PAGES;
			schema.items 	= getConfig.items;
			schema.page		= page;
			schema.hash		= null;
			schema.method 	= "cache";
			
			setAddress(schema.toURI);
		}
	}
	
	void previousItemHandler(SelectionEvent e)
	{
		Schema schema 	= Schema.parse(getAddress());
		uint count 		= m_browser.getCurrentItemCount;
		
		if (schema.mode == Schema.Mode.PAGES) {
			if (schema.page > 0) {
				schema.page = schema.page - 1;
				schema.method 	= "cache";
				
				setAddress(schema.toURI);
			}
		} else {
			uint page		= (count / schema.items);
			if (page > 0) {
				--page;
			}
			
			schema.mode 	= Schema.Mode.PAGES;
			schema.items 	= getConfig.items;
			schema.page		= page;
			schema.hash		= null;
			schema.method 	= "cache";
			
			setAddress(schema.toURI);
		}
	}
	
	void topItemHandler(SelectionEvent e)
	{
		Schema schema = Schema.parse(getAddress());
		
		schema.mode		= Schema.Mode.PAGES;
		schema.page		= 0;
		schema.items	= getConfig.items;
		schema.hash		= null;
		schema.method 	= "cache";
		
		setAddress(schema.toURI);
	}
	
	void configClickHandler(SelectionEvent e)
	{
		ToolItem	button	= cast(ToolItem)e.widget;
		ToolBar		bar		= button.getParent;
		Rectangle	rect	= button.getBounds;
		
		s_res_menu.setLocation(bar.toDisplay(rect.x, rect.y + rect.height));
		s_res_menu.setVisible(true);
	}
	
	class PostFinalizer: Runnable
	{
		ShingetsuBrowserPage parent;
		this (ShingetsuBrowserPage parent) { this.parent = parent; }
		void run()
		{
			if (parent !is null) {
				parent.refresh();
			}
		}
	}
	
	void postItemHandler(SelectionEvent e)
	{
		Schema 				schema	= Schema.parse(getAddress());
		BorwserPostWindow	child 	= new BorwserPostWindow;
		
		schema.mode 	= Schema.Mode.UNDEF;
		schema.method 	= "post";
		schema.hash		= m_target;
		
		child.setAddress(schema.toURI);
		child.setFinalizer(m_finalizer);
		child.run();
	}
	
	CBrowser getBrowser()
	{
		return m_browser;
	}
	
	void themeSelectionHandler(SelectionEvent e)
	{
		MenuItem item = cast(MenuItem)e.widget;
		getConfig.theme = item.getText();
		
		BrowserTabPage current = getCurrentTabPage();
		if (current !is null) {
			current.refresh();
		}
	}
	
	void changeItemsHandler(SelectionEvent e)
	{
		OneLineEditDialog 	dialog 	= new OneLineEditDialog("表示件数変更", "1ページの表示件数", 50);
		ToolItem			button	= cast(ToolItem)e.cData;
		ToolBar				bar		= button.getParent;
		Rectangle			rect	= button.getBounds;
		
		s_res_menu.setVisible(false);
		dialog.setText(std.string.toString(getConfig.items));
		if (dialog.run(bar.toDisplay(rect.x, rect.y + rect.height)) == DWT.OK) {
			try {
				getConfig.items = std.conv.toUint(dialog.getText());
				m_recent_item.setToolTipText(.std.string.format("最新%s件を表示", getConfig.items));
			} catch (Exception e) {
				MessageBox.showMessage(e.toString);
			}
		}
	}
	
	void createWidgets()
	{
		Composite page = getPage();
		page.setLayout(new GridLayout(2, false));
		ToolItem conf_item;
		
//		with (m_title = new TitleText(page, DWT.NONE)) {
//			setFontSize(10);
//			setMargin(5);
//			setBackground(new Color(null, 0xcc, 0xcc, 0xff));
//			setText("[Loading]");
//			setLayoutData(new GridData(GridData.FILL_HORIZONTAL));
//		}
		with (m_title = new CLabel(page, DWT.NONE)) {
			Color[3] 	gradient_colors;
			int[2] 		percent_array;
			
			percent_array[0] = 50;
			percent_array[1] = 100;
			
			gradient_colors[0] = Display.getDefault.getSystemColor(DWT.COLOR_TITLE_BACKGROUND);
			gradient_colors[1] = Display.getDefault.getSystemColor(DWT.COLOR_TITLE_BACKGROUND_GRADIENT);
			gradient_colors[2] = null;
			
			setFont(new Font(null, "MS UI Gothic", 10, DWT.BOLD));
			setBackground(gradient_colors, percent_array, false);
			setForeground(Display.getCurrent.getSystemColor(DWT.COLOR_TITLE_FOREGROUND));
			GridData grid_data = new GridData(GridData.FILL_HORIZONTAL);
			grid_data.grabExcessHorizontalSpace = true;
			grid_data.horizontalAlignment = GridData.FILL;
			grid_data.verticalAlignment = GridData.FILL;
			setLayoutData(grid_data);
		}
		
		Composite tool  = new Composite(page, DWT.NONE);
		tool.setLayout(new GridLayout(2, false));
		tool.setLayoutData(new GridData(GridData.END));
		
		ToolBar toolbar;
		with (toolbar = new ToolBar(tool, DWT.FLAT)) {
			setLayoutData(new GridData(GridData.BEGINNING));
		}
		
		new ToolItem(toolbar, DWT.SEPARATOR);
		with (conf_item = new ToolItem(toolbar, DWT.PUSH)) {
//			Image image = getIcon("cog");
//			image.transparentPixel = 0xffffff;	// white
//			image.setBackground(page.getBackground);
//			setImage(image);
			setText("設定");
			
			setToolTipText("設定");
			handleSelection(null, &configClickHandler);
		}
		new ToolItem(toolbar, DWT.SEPARATOR);
		with (new ToolItem(toolbar, DWT.PUSH)) {
//			Image image = getIcon("pencil_go");
//			image.transparentPixel = 0xffffff;	// white
//			image.setBackground(page.getBackground);
//			setImage(image);
			setText("投稿");
			
			setToolTipText("投稿");
			handleSelection(null, &postItemHandler);
		}
		new ToolItem(toolbar, DWT.SEPARATOR);
		with (new ToolItem(toolbar, DWT.PUSH)) {
			setText("1-");
			setToolTipText("１ページ目を表示");
			handleSelection(null, &topItemHandler);
		}
		with (new ToolItem(toolbar, DWT.PUSH)) {
			setText("前");
			setToolTipText("ひとつ前のページを表示");
			handleSelection(null, &previousItemHandler);
		}
		with (new ToolItem(toolbar, DWT.PUSH)) {
			setText("次");
			setToolTipText("次のページを表示");
			handleSelection(null, &nextItemHandler);
		}
		new ToolItem(toolbar, DWT.SEPARATOR);
		with (new ToolItem(toolbar, DWT.PUSH)) {
			setText("全件");
			setToolTipText("すべて表示");
			handleSelection(null, &allItemHandler);
		}
		new ToolItem(toolbar, DWT.SEPARATOR);
		with (m_recent_item = new ToolItem(toolbar, DWT.PUSH)) {
			setText("最新");
//			setToolTipText(std.string.format("最新%s件を表示", getConfig.items));
			setToolTipText(.std.string.format("最新%s件を表示", getConfig.items));
			handleSelection(null, &recentItemHandler);
		}
		
		m_browser = new ShingetsuBrowser(page, DWT.FLAT, this);
		m_browser.addLocationListener(new LocationChangeListener);
		m_browser.addStatusTextListener(new StatusChangeListener);
			
		GridData browser_grid = new GridData(GridData.FILL_BOTH);
		browser_grid.horizontalSpan = 3;
		m_browser.setLayoutData(browser_grid);
//		m_browser.drawLoadingPage(null);
		
		m_browser.addOpenWindowListener(new NewWindowListener(this));
		
		// thread menu
		char[][] themes;
		bool searchTheme(DirEntry *dirent)
		{
			if (dirent.isdir) {
				themes ~= std.path.getBaseName(dirent.name);
			}
			
			return true;
		}
		
		if (s_res_menu is null) {
			s_res_menu			= new Menu(getTabFolder.getShell());
			MenuItem 	item;
			Menu		sub_menu;
			
			with (item = new MenuItem(s_res_menu, DWT.CASCADE)) {
				setText("テーマ");
				setImage(getIcon("color_swatch"));
			}
			sub_menu = new Menu(item);
			item.setMenu(sub_menu);
			
			std.file.listdir(getConfig.templateDirectory, &searchTheme);
			foreach (e; themes) {
				item = new MenuItem(sub_menu, DWT.RADIO);
				item.setText(e);
				item.handleSelection(null, &themeSelectionHandler);
				
				if (std.string.icmp(e, getConfig.theme) == 0) {
					item.setSelection(1);
				}
			}
			
			item = new MenuItem(s_res_menu, DWT.SEPARATOR);
			with (item = new MenuItem(s_res_menu, DWT.PUSH)) {
				setText("表示件数の変更");
				setImage(getIcon("cog"));
				handleSelection(conf_item, &changeItemsHandler);
			}
		}
	}
	
	void setTitleDetail(char[] detail)
	{
		m_title.setText(std.string.format("(%s) %s", detail, m_title_text_origin));
	}
	
	void setTitle(char[] title)
	{
		m_title_text_origin = title;
		m_title.setText(title);
	}
	
	override void dispose()
	{
		if (m_browser !is null) {
			m_browser.dispose();
			m_browser = null;
			super.dispose();
		}
	}
	
	
public:
	this(char[] address, CTabFolder folder, int style)
	{
		super(address, folder, style);
		m_unique = time(null) ^ ((cast(uint)rand())>> 16);
		
		m_finalizer = new PostFinalizer(this);
	}
	
	TypeInfo getType()
	{
		return typeid(ShingetsuBrowserPage);
	}
	
	void setTitleText(char[] title)
	{
		setTitle(title);
		super.setTitleText(title);
	}
	
	int compareAddress(char[] address)
	{
		if (std.string.ifind(address, "shingetsu:") == 0
			&& std.string.ifind(getAddress, "shingetsu:") == 0)
		{
			Schema l_schema = Schema.parse(address);
			Schema r_schema = Schema.parse(getAddress);
			
			return std.string.cmp(l_schema.threadName, r_schema.threadName);
		}
		
		return std.string.cmp(address, getAddress);
	}
	
	void progressChanged(uint dummy = 0)
	{
		if (m_sw++ % 2) {
			setTitleImage(getIcon("transmit_blue"));
		} else {
			setTitleImage(getIcon("custom_transmit_dark_blue"));
		}
	}
	
	char[] getAddress()
	{
		return m_browser.getUrl();
	}
	
	void setAddress(char[] address)
	{
		m_browser.setUrl(address);
	}
	
	void refresh()
	{
		setAddress(getAddress());
	}
	
	void forward()
	{
		getBrowser.forward();
	}
	
	void back()
	{
		getBrowser.back();
	}
	
	void stop()
	{
		// do nothing
	}
	
	bool isStopEnabled()
	{
		return false;
	}
	
	bool isForwardEnabled()
	{
		return getBrowser.isForwardEnabled ? true: false;
	}
	
	bool isBackEnabled()
	{
		return getBrowser.isBackEnabled() ? true: false;
	}
}


// ウェブブラウザページ
class WebBrowserTabPage: BrowserTabPage
{
private:
	bool		m_disposed;

protected:
	CBrowser	m_browser;
	bool		m_busy;
	Menu		m_menu;
	
	bool isBusy()
	{
		return m_busy;
	}
	
	void busy(bool sw)
	{
		m_busy = sw;
	}
	
	// CBrowser Events
	
	class TitleChangeListener: TitleListener
	{
		void changed(TitleEvent e)
		{
			if (e.title.length != 0) {
				if (std.string.find(e.title, "://") != -1) {
					char[] title = std.string.split(e.title, "/")[length-1];
					if (title.length) {
						setTitleText(title);
					} else {
						setTitleText(e.title);
					}
				} else if (std.string.find(e.title, ":\\") != -1) {
					char[] title = std.string.split(e.title, "\\")[length-1];
					if (title.length) {
						setTitleText(title);
					} else {
						setTitleText(e.title);
					}
				} else {
					setTitleText(e.title);
				}
			} else {
				setTitleText("[EMPTY]");
			}
		}
	}
	
	class LocationChangeListener: LocationListener
	{
	public:
		void changing(LocationEvent e)
		{
			changeStatus();
		}
		
		void changed(LocationEvent e)
		{
			if (isSelective) {
				setAddressBarText(e.location);
				changeStatus();
			}
		}
	}
	
	class NewWindowListener: OpenWindowListener
	{
		public void open(WindowEvent e)
		{
			WebBrowserTabPage browser = new WebBrowserTabPage(CBrowser.ABOUT_BLANK, getTabFolder, getStyle);
			setPageParameter(browser);
			e.browser = browser.getBrowser();
		}
	}
	
	class StatusChangeListener: StatusTextListener
	{
		public void changed(StatusTextEvent e)
		{
			setStatusBarText(e.text);
		}
	}
	
	class LoadingListener: ProgressListener
	{
		uint sw = 0;
		public void changed(ProgressEvent e)
		{
			if (e.current != e.total) {
			if (sw++ % 2) {
					setTitleImage(getIcon("transmit_blue"));
				} else {
					setTitleImage(getIcon("custom_transmit_dark_blue"));
				}
				busy(true);
				changeStatus();
			} else {
				setTitleImage(getContentTypeIcon(std.path.getExt(getAddress())));
				sw = 0;
				busy(false);
				changeStatus();
			}
		}
		
		public void completed(ProgressEvent e) {
			// end
			setTitleImage(getContentTypeIcon(std.path.getExt(getAddress())));
			sw = 0;
			busy(false);
			changeStatus();
		}
	}
	
	// create widgets
	override void createWidgets()
	{
		Composite 	page = getPage();
		CTabItem	item = getTabItem();
		
		page.setLayout(new GridLayout(1, false));
		m_browser = new CBrowser(page, DWT.NONE);
		m_browser.addTitleListener(new TitleChangeListener);
		m_browser.addLocationListener(new LocationChangeListener);
		m_browser.addOpenWindowListener(new NewWindowListener);
		m_browser.addStatusTextListener(new StatusChangeListener);
		m_browser.addProgressListener(new LoadingListener);
		
		m_browser.setLayoutData(new GridData(GridData.FILL_BOTH));
		
		item.setText("[EMPTY]");
		
	}
	
	override void dispose()
	{
		if (!m_disposed) {
			m_browser.dispose();
			super.dispose();
			m_disposed = true;
		}
	}
	
public:
	this(char[] address, CTabFolder folder, int style)
	{
		super(address, folder, style);
		m_disposed = false;
	}
	
	override TypeInfo getType()
	{
		return typeid(WebBrowserTabPage);
	}
	
	override void setAddress(char[] address)
	{
		if (address.length == 0) {
			address = CBrowser.ABOUT_BLANK;
		} else if (std.string.find(address, "ttp://") == 0) {
			address = "h" ~ address;
		} else if (std.string.find(address, ":") == -1) {
			address = "http://" ~ address;
		}
		busy(true);
		m_browser.setUrl(address);
		changeStatus();
	}
	
	override char[] getAddress()
	{
		return m_browser.getUrl();
	}
	
	override void refresh()
	{
		busy(true);
		m_browser.refresh();
		changeStatus();
	}
	
	override void forward()
	{
		busy(true);
		m_browser.forward();
		changeStatus();
	}
	
	override void back()
	{
		busy(true);
		m_browser.back();
		changeStatus();
	}
	
	override bool isForwardEnabled()
	{
		return m_browser.isForwardEnabled() != 0;
	}
	
	override bool isBackEnabled()
	{
		return m_browser.isBackEnabled() != 0;
	}
	
	override void stop()
	{
		m_browser.stop();
	}
	
	override bool isStopEnabled()
	{
		return isBusy();
	}
	
	CBrowser getBrowser()
	{
		return m_browser;
	}
}


// タブブラウザクラス
class TabBrowser: ControlBase
{
protected:
	const char[]		DEFAULT_HOME_URI = "http://www.google.co.jp/";
	
	SashForm			m_vertical_sash;	// parent
	SashForm			m_horizontal_sash; 	// parent.parent
	CTabFolder 			m_tab_folder;
	Composite			m_parent;
	CCombo				m_address_bar;
	ToolItem			m_next;
	ToolItem			m_prev;
	ToolItem			m_stop;
	ToolItem			m_refresh;
	ToolItem			m_post;
	ToolItem			m_recent;
	ToolItem			m_list;
	ToolItem			m_new;
	ToolItem			m_visible_table;
	ToolItem			m_visible_tree;
	Menu				m_tab_menu;
	Menu				m_web_menu;
	Menu				m_add_favorite_menu;
	Menu				m_organize_menu;
	bool				m_focus_flag;
	char[][]			m_address;
	Label				m_status_bar;
	int					m_tree_percent;
	int					m_table_percent;
	Menu				m_home_menu;
	
	void appendAddress(char[] address)
	{
		if (address.length == 0) return;
		
		char[][] new_address;
		
		new_address ~= address;
		foreach (uri; m_address) {
			if (std.string.cmp(address, uri) != 0) {
				new_address ~= uri;
			}
		}
		delete m_address;
		m_address = new_address;
		
		m_address_bar.setItems(m_address);
		m_address_bar.select(0);
	}
	
	// アドレスバーEnter
	class CommandListener: KeyListener
	{
		public void keyReleased(KeyEvent e)
		{
			
		}
		
		public void keyPressed(KeyEvent e)
		{
			if(e.character == DWT.CR) {
				char[]	address = m_address_bar.getText;
				
				if (address.length == 0) {
					return;
				}
				if (std.string.find(address, "ttp://") == 0) {
					address = "h" ~ address;
				}
				if (std.string.find(address, ":") == -1) {
					address = "http://" ~ address;
				}
				
				addPage(address);
				appendAddress(address);
			}
		}
	}
	
	class LoadingListener: ProgressListener
	{
		public void changed(ProgressEvent e)
		{
			// とりあえず呼ぶ
			updateToolItems();
		}
		
		public void completed(ProgressEvent e)
		{
			// とりあえず呼ぶ
			updateToolItems();
		}
	}
	
	void updateToolItems()
	{
		BrowserTabPage page	= getCurrentPage();
		
		if (page is null) {
			m_next.setEnabled(false);
			m_prev.setEnabled(false);
			m_stop.setEnabled(false);
			
			return;
		}
		
		if (page.isForwardEnabled()) {
			m_next.setEnabled(true);
		} else {
			m_next.setEnabled(false);
		}
		if (page.isBackEnabled()) {
			m_prev.setEnabled(true);
		} else {
			m_prev.setEnabled(false);
		}
		if (page.isStopEnabled()) {
			m_stop.setEnabled(true);
			m_refresh.setEnabled(false);
		} else {
			m_stop.setEnabled(false);
			m_refresh.setEnabled(true);
		}
	}
	
	void hideTreeHandler(SelectionEvent e)
	{
		ToolItem item = cast(ToolItem)e.cData;
		Control[] controls = m_horizontal_sash.getControls(0);
		if (item.getSelection) {
			int[2] weights;
			weights[0] = m_tree_percent;
			weights[1] = 100 - m_tree_percent;
			controls[0].setVisible(true);
			m_horizontal_sash.setWeights(weights);
			
			item.setToolTipText("サイドウィンドウを隠す");
		} else {
			int[] weights 	= m_horizontal_sash.getWeights();
			int percent 	= 0;
			
			m_tree_percent	= cast(int)(cast(double)weights[0] / (weights[0] + weights[1]) * 100);
			
			weights[0] 		= percent;
			weights[1] 		= 100 - percent;
			m_horizontal_sash.setWeights(weights);
			controls[0].setVisible(false);
			
			item.setToolTipText("サイドウィンドウを表示");
		}
	}
	
	void hideTableHandler(SelectionEvent e)
	{
		ToolItem item = cast(ToolItem)e.cData;
		Control[] controls = m_vertical_sash.getControls(0);
		if (item.getSelection) {
			int[2] weights;
			weights[0] = m_table_percent;
			weights[1] = 100 - m_table_percent;
			controls[0].setVisible(true);
			m_recent.setEnabled(true);
			m_list.setEnabled(true);
			m_new.setEnabled(true);
			m_vertical_sash.setWeights(weights);
			
			item.setToolTipText("トップウィンドウを隠す");
		} else {
			int[] weights 	= m_vertical_sash.getWeights();
			int percent 	= 0;
			
			m_table_percent	= cast(int)(cast(double)weights[0] / (weights[0] + weights[1]) * 100);
			
			weights[0] 		= percent;
			weights[1] 		= 100 - percent;
			
			m_vertical_sash.setWeights(weights);
			controls[0].setVisible(false);
			m_recent.setEnabled(false);
			m_list.setEnabled(false);
			m_new.setEnabled(false);
			
			item.setToolTipText("トップウィンドウを表示");
		}
	}
	
	void disposeHandler(DisposeEvent e)
	{
	}
	
	void deserialize()
	{
		int[2] weights;
		int table_percent, tree_percent;
		char[] address;
		
		try {
			VArray array = getGUIConfig();
			
			if (array.toLong == 3) {
				table_percent	= array[0].toLong;
				tree_percent	= array[1].toLong;
				address			= array[2].toString;
				
				weights[0] = table_percent;
				weights[1] = 100 - table_percent;
				m_vertical_sash.setWeights(weights);
				
				weights[0] = tree_percent;
				weights[1] = 100 - tree_percent;
				m_horizontal_sash.setWeights(weights);
				
				m_address = std.string.split(address, "\n");
				if (m_address.length != 0) {
					m_address_bar.setItems(m_address);
				}
			}
		} catch (Exception) {
			weights[0] = 20;
			weights[1] = 80;
			m_horizontal_sash.setWeights(weights);
			weights[0] = 25;
			weights[1] = 75;
			m_vertical_sash.setWeights(weights);
		}
	}
	
	void serialize()
	{
		char[] address = std.string.join(m_address, "\n");
		long table_percent, tree_percent;
		
		if (m_visible_table.getSelection) {
			int[] weights 	= m_vertical_sash.getWeights();
			table_percent = cast(int)(cast(double)weights[0] / (weights[0] + weights[1]) * 100);
		} else {
			table_percent = m_table_percent;
		}
		
		if (m_visible_tree.getSelection) {
			int[] weights 	= m_horizontal_sash.getWeights();
			tree_percent = cast(int)(cast(double)weights[0] / (weights[0] + weights[1]) * 100);
		} else {
			tree_percent = m_tree_percent;
		}
		
		VArray array = VArray.make(
			table_percent,
			tree_percent,
			address
		);
		setGUIConfig(array);
	}
	
	// menu events
	void menuPopHandler(MouseEvent e)
	{
		if (e.button != 3) {
			return; // clicked button is not right button
		}
		BrowserTabPage page = getCurrentPage();
		
		if (page.getType is typeid(WebBrowserTabPage)) {
			m_web_menu.setLocation(m_tab_folder.toDisplay(e.x, e.y));
			m_web_menu.setVisible(true);
		} else {
			m_tab_menu.setLocation(m_tab_folder.toDisplay(e.x, e.y));
			m_tab_menu.setVisible(true);
		}
	}
	
	void threadCloseHandler(SelectionEvent e)
	{
		CTabItem	item = m_tab_folder.getSelection();
		if (item !is null) {
			item.dispose();
		}
	}
	
	void threadCloseAllHandler(SelectionEvent e)
	{
		foreach (item; m_tab_folder.getItems) {
			if (item !is null) {
				item.dispose();
				item = null;
			}
		}
	}
	
	void threadGatewayOpenHandler(SelectionEvent e)
	{
		BrowserTabPage page = getCurrentPage();
		char[] title = page.getTitleText();
		Configuration config 	= getConfig();
		char[] url = std.string.format("http://%s:%s/thread.cgi/%s", 
						config.nodeName, config.nodePort, std.uri.encode(title));
		
		addPage(url);
	}
	
	void titleCopyHandler(SelectionEvent e)
	{
		BrowserTabPage page = getCurrentPage();
		char[] title = page.getTitleText();
		
		copyText(title);
	}
	
	void shingetsuURLCopyHandler(SelectionEvent e)
	{
		BrowserTabPage page = getCurrentPage();
		char[] title = page.getTitleText();
		
		copyText(std.string.format("[[%s]]", title));
	}
	
	void hovenURLCopyHandler(SelectionEvent e)
	{
		BrowserTabPage page = getCurrentPage();
		char[] title = page.getTitleText();
		Schema schema 	= new Schema(title);
		schema.method 	= "get";
		
		copyText(std.string.format(schema.toURI));
	}
	
	void gatewayURLCopyHandler(SelectionEvent e)
	{
		BrowserTabPage page = getCurrentPage();
		char[] title = page.getTitleText();
		Configuration config = getConfig();
		char[] url = std.string.format("http://%s:%s/thread.cgi/%s", 
						config.nodeName, config.nodePort, std.uri.encode(title));
		
		copyText(url);
	}
	
	void gatewayTitleAndURLCopyHandler(SelectionEvent e)
	{
		BrowserTabPage page = getCurrentPage();
		char[] title = page.getTitleText();
		Configuration config = getConfig();
		char[] url = std.string.format("http://%s:%s/thread.cgi/%s", 
						config.nodeName, config.nodePort, std.uri.encode(title));
		
		copyText(title ~ "\x0d\x0a" ~ url);
	}
	
	void webTitleAndURLCopyHandler(SelectionEvent e)
	{
		BrowserTabPage page = getCurrentPage();
		copyText(page.getTitleText ~ "\x0d\x0a" ~ page.getAddress);
	}
	
	void webURLCopyHandler(SelectionEvent e)
	{
		BrowserTabPage page = getCurrentPage();
		
		copyText(page.getAddress);
	}
	
public:
	this(Composite parent)
	{
		m_parent = new Composite(parent, DWT.FLAT);
		m_vertical_sash 	= cast(SashForm)parent; 		// うぁああ
		m_horizontal_sash 	= cast(SashForm)parent.parent;	// うぁああ
		
		createWidgets();
		deserialize();
	}
	
	void finalize()
	{
		serialize();
	}
	
	void createWidgets()
	{
		m_parent.setLayout(new GridLayout(7, false));
		
		GridData top_bar 	= new GridData(GridData.HORIZONTAL_ALIGN_FILL);
		GridData group_grid = new GridData(GridData.BEGINNING);
		GridData a_grid 	= new GridData(GridData.BEGINNING);
		GridData b_grid 	= new GridData(GridData.BEGINNING);
		GridData d_grid 	= new GridData(GridData.BEGINNING);
		GridData f_grid 	= new GridData(GridData.BEGINNING);
		GridData g_grid 	= new GridData(GridData.BEGINNING);
		GridData i_grid 	= new GridData(GridData.BEGINNING);
		GridData r_grid 	= new GridData(GridData.BEGINNING);
		GridData new_grid 	= new GridData(GridData.BEGINNING);
		GridData prev_grid 	= new GridData(GridData.BEGINNING);
		GridData next_grid 	= new GridData(GridData.BEGINNING);
		GridData home_grid 	= new GridData(GridData.BEGINNING);
		GridData uri_grid 	= new GridData(GridData.FILL_HORIZONTAL);
		GridData tab_grid 	= new GridData(GridData.FILL_BOTH);
		
		// ブラウザ操作ツールバー
		ToolBar toolbar2 = new ToolBar(m_parent, DWT.FLAT);
		GridData tool_grid2 = new GridData(GridData.BEGINNING);
		tool_grid2.horizontalSpan = 2;
		toolbar2.setLayoutData(tool_grid2);
		
		m_recent = new ToolItem(toolbar2, DWT.RADIO);
		m_recent.setText("新着");
		m_recent.setToolTipText("新着スレッド一覧を表示");
		
		m_list = new ToolItem(toolbar2, DWT.RADIO);
		m_list.setText("一覧");
		m_list.setToolTipText("購読中スレッド一覧を表示");
		
		cast(void)new ToolItem(toolbar2, DWT.SEPARATOR);
		
		m_new = new ToolItem(toolbar2, DWT.PUSH);
		m_new.setText("新規");
		m_new.handleSelection(m_new, &newThreadHandler);
		m_new.setToolTipText("新規スレッドを作成");
		
		cast(void)new ToolItem(toolbar2, DWT.SEPARATOR);
		
		ToolBar toolbar = new ToolBar(m_parent, DWT.FLAT);
		GridData tool_grid = new GridData(GridData.BEGINNING);
		tool_grid.horizontalSpan = 3;
		toolbar.setLayoutData(tool_grid);
		
		with (m_visible_tree = new ToolItem(toolbar, DWT.CHECK)) {
			setImage(getIcon("folder"));
			handleSelection(m_visible_tree, &hideTreeHandler);
			setSelection(1);
			setToolTipText("サイドウィンドウを隠す");
		}
		
		with (m_visible_table = new ToolItem(toolbar, DWT.CHECK)) {
			setImage(getIcon("table"));
			handleSelection(m_visible_table, &hideTableHandler);
			setSelection(1);
			setToolTipText("トップウインドウを隠す");
		}
		cast(void)new ToolItem(toolbar, DWT.SEPARATOR);
/*		
		ToolItem newb = new ToolItem(toolbar, DWT.PUSH);
		Image nicon = getIcon("page_add");
		nicon.transparentPixel = 0xffffff;	// white
		nicon.setBackground(toolbar.getBackground());
		newb.setImage(nicon);
		newb.setToolTipText("新規ページを開く");
		newb.handleEvent(null, DWT.Selection, &newClickHandler);
*/		
		ToolItem deleteb = new ToolItem(toolbar, DWT.PUSH);
		Image dicon = getIcon("page_delete");
		dicon.transparentPixel = 0xffffff;	// white
		dicon.setBackground(toolbar.getBackground());
		deleteb.setImage(dicon);
		deleteb.setToolTipText("現在開いているページを閉じる");
		
		deleteb.handleEvent(null, DWT.Selection, &deleteClickHandler);
		
		cast(void)new ToolItem(toolbar, DWT.SEPARATOR);
		
		ToolItem saku = new ToolItem(toolbar, DWT.PUSH);
		saku.setImage(getIcon("saku"));
		saku.handleEvent(null, DWT.Selection, &sakuClickHandler);
		saku.setToolTipText("ゲートウェイを開く");
		
		cast(void)new ToolItem(toolbar, DWT.SEPARATOR);
		
		ToolItem favorite = new ToolItem(toolbar, DWT.DROP_DOWN);
		favorite.setImage(getIcon("icon_favourites"));
		favorite.handleSelection(favorite, &favoriteClickHandler);
		favorite.setToolTipText("現在開いているページをお気に入りへ登録");
		
		ToolItem home = new ToolItem(toolbar, DWT.DROP_DOWN);
		home.setImage(getIcon("icon_home"));
		home.handleEvent(null, DWT.Selection, &homeClickHandler);
		home.setToolTipText("ホームを開く");
		
		cast(void)new ToolItem(toolbar, DWT.SEPARATOR);
		
		m_prev 	= new ToolItem(toolbar, DWT.PUSH);
		m_prev.setImage(getIcon("arrow_left"));
		m_prev.handleEvent(null, DWT.Selection, &prevClickHandler);
		m_prev.setToolTipText("ひとつ前のページへ戻る");
		
		m_next	= new ToolItem(toolbar, DWT.PUSH);
		m_next.setImage(getIcon("arrow_right"));
		m_next.handleEvent(null, DWT.Selection, &nextClickHandler);
		m_next.setToolTipText("ひとつ先のページへ進む");
		
		m_refresh = new ToolItem(toolbar, DWT.PUSH);
		m_refresh.setImage(getIcon("action_refresh"));
		m_refresh.handleEvent(null, DWT.Selection, &refreshClickHandler);
		m_refresh.setToolTipText("現在開いているページを更新");
		
		m_stop = new ToolItem(toolbar, DWT.PUSH);
		m_stop.setImage(getIcon("action_stop"));
		m_stop.handleEvent(null, DWT.Selection, &stopClickHandler);
		m_stop.setToolTipText("現在開こうとしているページの通信処理を中止");
		
		m_address_bar  	= new CCombo(m_parent, DWT.FLAT|DWT.DROP_DOWN);
		uri_grid.horizontalSpan = 2;
		m_address_bar.setLayoutData(uri_grid);
		m_address_bar.addKeyListener(new CommandListener);
		
//		m_address_bar.handleFocusIn(null, &addressBarFocusInHandler);
		m_address_bar.text.handleFocusOut(null, &addressBarFocusOutHandler);
		m_address_bar.text.handleMouseDown(null, &addressBarClickHandler);
		
		// タブブラウザ
		with (m_tab_folder = new CTabFolder(m_parent, DWT.TOP | DWT.FLAT |DWT.BORDER | DWT.CLOSE))
		{
			Color[3] 	gradient_colors;
			int[2] 		percent_array;
			
			percent_array[0] = 50;
			percent_array[1] = 100;
			
			gradient_colors[0] = Display.getDefault.getSystemColor(DWT.COLOR_TITLE_BACKGROUND);
			gradient_colors[1] = Display.getDefault.getSystemColor(DWT.COLOR_TITLE_BACKGROUND_GRADIENT);
			gradient_colors[2] = null;
			
			setTabHeight(22);
			setSelection(0);
			setSimple(false);
			setBackground(Display.getDefault.getSystemColor(DWT.COLOR_LIST_BACKGROUND));
			setForeground(Display.getDefault.getSystemColor(DWT.COLOR_LIST_FOREGROUND));
			setSelectionForeground(Display.getDefault.getSystemColor(DWT.COLOR_TITLE_FOREGROUND));
			setSelectionBackground(gradient_colors, percent_array, true);
			handleSelection(m_tab_folder, &tabSelectionHandler);
			
			tab_grid.horizontalSpan = 7;
			
			m_tab_folder.setLayoutData(tab_grid);
		}
		
		GridData status_bar = new GridData(GridData.HORIZONTAL_ALIGN_FILL);
		m_status_bar = new Label(m_parent, DWT.FLAT);
		m_status_bar.setForeground = m_parent.getDisplay().getSystemColor(DWT.COLOR_BLUE);
		
		status_bar.horizontalSpan = 7;
		m_status_bar.setLayoutData(status_bar);
		
		
		// shingetsu menu
		m_tab_menu	= new Menu(m_tab_folder.getShell());
		with (new MenuItem(m_tab_menu, DWT.PUSH)) {
			setText("閉じる");
			setImage(getIcon("page_delete"));
			handleSelection(null, &threadCloseHandler);
		}
		cast(void)new MenuItem(m_tab_menu, DWT.SEPARATOR);
		with (new MenuItem(m_tab_menu, DWT.PUSH)) {
			setText("タブを全て閉じる");
			setImage(getIcon("delete"));
			handleSelection(null, &threadCloseAllHandler);
		}
		
		cast(void)new MenuItem(m_tab_menu, DWT.SEPARATOR);
		
		with (new MenuItem(m_tab_menu, DWT.PUSH)) {
			setText("ゲートウェイを開く");
			setImage(getIcon("monitor"));
			handleSelection(null, &threadGatewayOpenHandler);
		}
		
		cast(void)new MenuItem(m_tab_menu, DWT.SEPARATOR);
		
		with (new MenuItem(m_tab_menu, DWT.PUSH)) {
			setText("ブックマークへ追加");
			setImage(getIcon("star"));
			handleSelection(null, &addBookmarkHandler);
		}
		
		cast(void)new MenuItem(m_tab_menu, DWT.SEPARATOR);
		
		with (new MenuItem(m_tab_menu, DWT.PUSH)) {
			setText("タイトルをコピー");
			handleSelection(null, &titleCopyHandler);
		}
		with (new MenuItem(m_tab_menu, DWT.PUSH)) {
			setText("URLを新月形式でコピー");
			handleSelection(null, &shingetsuURLCopyHandler);
		}
		with (new MenuItem(m_tab_menu, DWT.PUSH)) {
			setText("ゲートウェイのURLをコピー");
			handleSelection(null, &gatewayURLCopyHandler);
		}
		with (new MenuItem(m_tab_menu, DWT.PUSH)) {
			setText("タイトルとゲートウェイのURLをコピー");
			handleSelection(null, &gatewayTitleAndURLCopyHandler);
		}
		with (new MenuItem(m_tab_menu, DWT.PUSH)) {
			setText("URLをHoven形式でコピー");
			handleSelection(null, &hovenURLCopyHandler);
		}
		
		// web menu
		m_web_menu	= new Menu(m_tab_folder.getShell());
		with (new MenuItem(m_web_menu, DWT.PUSH)) {
			setText("閉じる");
			setImage(getIcon("page_delete"));
			handleSelection(null, &threadCloseHandler);
		}
		
		cast(void)new MenuItem(m_web_menu, DWT.SEPARATOR);
		with (new MenuItem(m_web_menu, DWT.PUSH)) {
			setText("タブを全て閉じる");
			setImage(getIcon("delete"));
			handleSelection(null, &threadCloseAllHandler);
		}
		cast(void)new MenuItem(m_web_menu, DWT.SEPARATOR);
		
		with (new MenuItem(m_web_menu, DWT.PUSH)) {
			setText("ブックマークへ追加");
			setImage(getIcon("star"));
			handleSelection(null, &addBookmarkHandler);
		}
		with (new MenuItem(m_web_menu, DWT.PUSH)) {
			setText("IEのお気に入りへ追加");
			setImage(getIcon("folder_star"));
			handleSelection(null, &addIEFavoriteHandler);
		}
		
		cast(void)new MenuItem(m_web_menu, DWT.SEPARATOR);
		
		with (new MenuItem(m_web_menu, DWT.PUSH)) {
			setText("タイトルとURLをコピー");
			handleSelection(null, &webTitleAndURLCopyHandler);
		}
		with (new MenuItem(m_web_menu, DWT.PUSH)) {
			setText("URLをコピー");
			handleSelection(null, &webURLCopyHandler);
		}
		m_tab_folder.handleMouseUp(null, &menuPopHandler);
		
		// add favorite menu
		m_add_favorite_menu = new Menu(m_tab_folder.getShell());
		with (new MenuItem(m_add_favorite_menu, DWT.PUSH)) {
			setText("ブックマークへ追加");
			setImage(getIcon("star"));
			handleSelection(null, &addBookmarkHandler);
		}
		with (new MenuItem(m_add_favorite_menu, DWT.PUSH)) {
			setText("IEのお気に入りへ追加");
			setImage(getIcon("folder_star"));
			handleSelection(null, &addIEFavoriteHandler);
		}
		
		// organize favorites menu
		m_organize_menu = new Menu(m_tab_folder.getShell());
		with (new MenuItem(m_organize_menu, DWT.PUSH)) {
			setText("ブックマークを整理");
			setImage(getIcon("folder"));
			handleSelection(null, &organizeBookmarkHandler);
		}
		with (new MenuItem(m_organize_menu, DWT.PUSH)) {
			setText("IEのお気に入りを整理");
			setImage(getIcon("folder_star"));
			handleSelection(null, &organizeIEFavoriteHandler);
		}
		
		updateToolItems();
	}
	
	BrowserTabPage addPage(char[] address)
	{
		BrowserTabPage page	= createTabPage(address, m_tab_folder, DWT.NONE);
		setPageParameter(page);
		page.setAddress(address);
		
		return page;
	}
	
	void setPageParameter(BrowserTabPage page)
	{
		page.setData("ADDRESS", new StringObj(page.getAddress)); // URI
		page.getTabItem.setData(page);
		page.setAddressBar(m_address_bar);
		page.setStatusBar(m_status_bar);
		
		page.handleSetParameter(&setPageParameter);
		page.handleStatusChanged(&updateToolItems);
	}
	
	BrowserTabPage changePage(char[] address)
	{
		int count = m_tab_folder.getItemCount;
		
		foreach (item; m_tab_folder.getItems) {
			BrowserTabPage page = cast(BrowserTabPage)item.getData();
			
			if (page.compareAddress(address) == 0) {
				m_tab_folder.setSelection(item);
				page.refresh();
				updateToolItems();
				
				return page;
			}
		}
		return addPage(address);
	}
	
	BrowserTabPage getCurrentPage()
	{
		CTabItem item = m_tab_folder.getSelection();
		
		if (item is null) {
			return null;
		}
		
		return cast(BrowserTabPage)item.getData();
	}
	
	void addressBarFocusInHandler(FocusEvent e)
	{
//		m_address_bar.text.setFocus();
	}
	
	void addressBarFocusOutHandler(FocusEvent e)
	{
		m_focus_flag = false;
	}
	
	void addressBarClickHandler(MouseEvent e)
	{
		if (! m_focus_flag) {
			m_address_bar.parent.setFocus();
			m_address_bar.text.setFocus();
			m_focus_flag = true;
		}
	}
	
	void tabSelectionHandler(SelectionEvent e)
	{
		BrowserTabPage page = getCurrentPage();
		
		if (page !is null) {
			m_address_bar.setText(page.getAddress);
			updateToolItems();
		}
	}
	
	void sakuClickHandler(Event e)
	{
		Configuration config = getConfig;
		char[] gateway;
		
		if (config.nodePort != 80) {
			gateway = std.string.format("http://%s:%s/", config.nodeName, config.nodePort);
		} else {
			gateway = std.string.format("http://%s/", getConfig.nodeName);
		}
		
		addPage(gateway);
	}
	
	void nextClickHandler(Event e)
	{
		BrowserTabPage page = getCurrentPage();
		
		if (page !is null) {
			page.forward();
		}
	}
	
	void prevClickHandler(Event e)
	{
		BrowserTabPage page = getCurrentPage();
		
		if (page !is null) {
			page.back();
		}
	}
	
	class ThreadRegister: Runnable
	{
		char[] title;
		TabBrowser parent;
		
		this(char[] title, TabBrowser parent)
		{
			this.title = title;
			this.parent = parent;
		}
		
		int newThread()
		{
			ThreadIO io = new ThreadIO(new InternetAddress(getConfig.nodeName, getConfig.nodePort));
			io.registerThread(title);
			
			return 0;
		}
		
		void run()
		{
/*
			Thread thread = new Thread(&newThread);
			Shell shell = m_parent.getShell();
			Display display = shell.getDisplay();
			
			try {
				thread.start();
				while (! shell.isDisposed()) {
					if (!display.readAndDispatch()) {
						if (thread.getState == Thread.TS.TERMINATED) {
							break;
						}
						Sleep(10);
					}
				}
			} catch (Object e) {
				MessageBox.showMessage(e.toString, getVersionString(), shell);
			}
			thread.wait();
			
			if (m_list.getSelection) {
				m_list.sendEvent(DWT.Selection);
			} else {
				m_recent.sendEvent(DWT.Selection);
			}
*/
			Schema schema 	= new Schema(title);
			schema.method 	= "get";
			schema.mode 	= Schema.Mode.BACK_ROWS;
			schema.items	= getConfig.items;
			parent.addPage(schema.toURI);
		};
	}
	
	void newThreadHandler(SelectionEvent e)
	{
		char[] title;
		OneLineEditDialog 	dialog 	= new NewThreadEditDialog("新規スレッド作成", "スレッド名", 256);
		ToolItem			button	= cast(ToolItem)e.cData;
		ToolBar				bar		= button.getParent;
		Rectangle			rect	= button.getBounds;
		
		dialog.setButtonText("作成");
		if (dialog.run(bar.toDisplay(rect.x, rect.y + rect.height)) == DWT.OK) {
			title = dialog.getText();
			
			if (title.length != 0) {
				BusyIndicator.showWhile(null, new ThreadRegister(title, this));
			}
		}
	}
	
	void newClickHandler(Event e)
	{
		addPage(CBrowser.ABOUT_BLANK);
	}
	
	void deleteClickHandler(Event e)
	{
		CTabItem item = m_tab_folder.getSelection();
		if (item !is null) {
			item.dispose();
		}
	}
	
	void refreshClickHandler(Event e)
	{
		BrowserTabPage page = getCurrentPage();
		
		if (page !is null) {
			page.refresh();
		}
	}
	
	void stopClickHandler(Event e)
	{
		BrowserTabPage page = getCurrentPage();
		
		if (page !is null) {
			page.stop();
		}
	}
	
	void editHome(SelectionEvent e)
	{
		OneLineEditDialog 	dialog 	= new OneLineEditDialog("ホームページ変更", "アドレス", 256);
		ToolItem			button	= cast(ToolItem)e.cData;
		ToolBar				bar		= button.getParent;
		Rectangle			rect	= button.getBounds;
		
		m_home_menu.setVisible(false);
		dialog.setText(getConfig.homeURL);
		if (dialog.run(bar.toDisplay(rect.x, rect.y + rect.height)) == DWT.OK) {
			getConfig.homeURL = dialog.getText();
		}
	}
	
	void addIEFavoriteHandler(SelectionEvent e)
	{
		BrowserTabPage page = getCurrentPage();
		if (page is null) {
			return;
		}
		addFavorite(m_parent.getShell(), getSystemFavoriteDirectory(), page.getAddress(), page.getTitleText);
	}
	
	void addBookmarkHandler(SelectionEvent e)
	{
		BrowserTabPage page = getCurrentPage();
		if (page is null) {
			return;
		}
		addFavorite(m_parent.getShell(), getConfig.favoritesDirectory, page.getAddress(), page.getTitleText);
	}
	
	void organizeBookmarkHandler(SelectionEvent e)
	{
		organizeFavorite(m_parent.getShell(), getConfig.favoritesDirectory);
	}
	
	void organizeIEFavoriteHandler(SelectionEvent e)
	{
		organizeFavorite(m_parent.getShell(), getSystemFavoriteDirectory());
	}
	
	void favoriteClickHandler(SelectionEvent e)
	{
		ToolItem	button	= cast(ToolItem)e.widget;
		ToolBar		bar		= button.getParent;
		Rectangle	rect	= button.getBounds;
		
		if (e.detail == DWT.ARROW) {
			m_organize_menu.setLocation(bar.toDisplay(rect.x, rect.y + rect.height));
			m_organize_menu.setVisible(true);
		} else {
			BrowserTabPage page = getCurrentPage();
			if (page is null) {
				return;
			}
			
			if (page.getType is typeid(WebBrowserTabPage)) {
				m_add_favorite_menu.setLocation(bar.toDisplay(rect.x, rect.y + rect.height));
				m_add_favorite_menu.setVisible(true);
			} else {
				addFavorite(m_parent.getShell(), getConfig.favoritesDirectory, page.getAddress(), page.getTitleText);
			}
		}
	}
	
	void homeClickHandler(Event e)
	{
		char[] home = getConfig.homeURL;
		
		if (e.detail == DWT.ARROW) {
			if (m_home_menu is null) {
				m_home_menu			= new Menu(m_tab_folder.getShell());
				MenuItem 	item	= new MenuItem(m_home_menu, DWT.NONE);
				item.setText("変更");
				item.setImage(getIcon("cog"));
				item.handleSelection(e.widget, &editHome); // first widget
			}
			ToolItem	button	= cast(ToolItem)e.widget;
			ToolBar		bar		= button.getParent;
			Rectangle	rect	= button.getBounds;
			
			m_home_menu.setLocation(bar.toDisplay(rect.x, rect.y + rect.height));
			m_home_menu.setVisible(true);
		} else {
			addPage(home);
		}
	}
	
	void handleRecentClick(Object o, void delegate(SelectionEvent) func)
	{
		m_recent.handleSelection(o, func);
	}
	
	void handleListClick(Object o, void delegate(SelectionEvent) func)
	{
		m_list.handleSelection(o, func);
	}
	
	void handleNewClick(Object o, void delegate(Event) func)
	{
		m_new.handleEvent(o, DWT.Selection, func);
	}
}


// スレッドアイテム
class ThreadItem
{
	enum SortIndex: int {
		NO			= 0,	// no
		TITLE,				// スレッドタイトル
		THREADSIZE,			// スレッドサイズ
		ENERGY,				// 活発度
		REC,				// 投稿数
		LOCALREC,			// 既読数
		MODDATE,			// 更新日
	}
	static SortIndex 	sortIndex;
	static int			sortDirection; // DWT.UP or DWT.DOWN
	
	char[]	no;
	char[]	title;
	char[]	threadSize;
	char[]	energy;
	char[]	records;
	char[]	localRecords;
	char[]	lastModifiedTime;
	
	int		noNum;
	int		threadSizeNum;
	int		energyNum;
	int		recordsNum;
	int		localRecordsNum;
	time_t	lastModifiedTimeNum;
	char[]	path;
	
	this(int n, ThreadResponseEvent event)
	{
		noNum				= n;
		no					= std.string.format("%04d", noNum);
		title				= event.title.dup;
		threadSizeNum		= event.threadSize;
		threadSize			= fileSizeString(threadSizeNum);
		recordsNum			= event.records;
		records				= std.string.toString(recordsNum);
//		energy		= std.string.toString(0);
//		localRecords		= std.string.toString(0);
		
		lastModifiedTimeNum	= event.lastModifiedTime;
		lastModifiedTime	= dateString(lastModifiedTimeNum + 3600 * 16);
		path				= event.path.dup;
	}
	
	char[] dateString(time_t ltime)
	{
		tm vtm = getLocalTime(&ltime);
		
		return std.string.format("%04d %02d/%02d %02d:%02d:%02d", 
								(vtm.tm_year + 1900), vtm.tm_mon + 1, vtm.tm_mday,
								vtm.tm_hour, vtm.tm_min, vtm.tm_sec);
	}
	
	static char[] fileSizeString(uint file_size)
	{
		double estimate;
		char[] unit;
		
		if (file_size < 1024) {
			estimate	= file_size;
			unit     	= "  B";
		} else if (file_size < 1024 * 1000) {
			estimate	= cast(double)file_size / 1024;
			unit		= " KB";
		} else if (file_size < 1024 * 1000 * 1000) {
			estimate	= cast(double)file_size / (1024 * 1000);
			unit		= " MB";
		} else {
			estimate	= cast(double)file_size / (1024 * 1000 * 1000);
			unit		= " GB";
		}
		
		return std.string.format("%.1f%s", estimate, unit);
	}
	
	void dispose()
	{
		delete no;
		delete title;
		delete threadSize;
		delete path;
		delete lastModifiedTime;
		delete records;
//		delete energy;
//		delete localRecords;
		
		no = null;
		title = null;
		threadSize = null;
		path = null;
		lastModifiedTime = null;
		records = null;
//		energy = null;
//		localRecords = null;
	}
	
	int opCmp(Object o)
	{
		ThreadItem rhs = cast(ThreadItem)o;
		int cmp;
		
		switch (sortIndex) {
		case SortIndex.NO:
			cmp = rhs.noNum - noNum;
			break;
		case SortIndex.TITLE:
			cmp = std.string.cmp(rhs.title, title);
			break;
		case SortIndex.THREADSIZE:
			cmp = rhs.threadSizeNum - threadSizeNum;
			break;
		case SortIndex.ENERGY:
			cmp = rhs.energyNum - energyNum;
			break;
		case SortIndex.REC:
			cmp = rhs.recordsNum - recordsNum;
			break;
		case SortIndex.LOCALREC:
			cmp = rhs.localRecordsNum - localRecordsNum;
			break;
		case SortIndex.MODDATE:
			if (rhs.lastModifiedTimeNum == lastModifiedTimeNum) {
				cmp = 0;
			} else if (rhs.lastModifiedTimeNum > lastModifiedTimeNum) {
				cmp = 1;
			} else {
				cmp = -1;
			}
			break;
		default:
			assert(0);
		}
		
		if (cmp == 0 && sortIndex != SortIndex.MODDATE) {
			if (rhs.lastModifiedTimeNum == lastModifiedTimeNum) {
				cmp = 0;
			} else if (rhs.lastModifiedTimeNum > lastModifiedTimeNum) {
				cmp = 1;
			} else {
				cmp = -1;
			}
		}
		if (sortDirection == DWT.UP) {
			cmp *= -1;
		}
		
		return cmp;
	}
}

char[][] getTags(char[] s)
{
	char[][] tags;
	int perr = 0, perl = 0;
	while ((perl = std.string.find(s, "【")) != -1) {
		s = s[perl + 3 .. length];
		perr = std.string.find(s, "】");
		if (perr == -1) {
			break;
		}
		tags ~= s[0 .. perr];
		s = s[perr + 2 .. length];
	}
	
	return tags;
}

// スレッド一覧テーブル
class ThreadTable: ControlBase
{
protected:
	alias WrapperObject!(ThreadItem.SortIndex) SortIndexWrapper;
	
	bool				m_list_selection;
	CCombo				m_tag_filter;
	Text				m_pattern_filter;
	Table				m_table;
	int					m_style;
	Composite			m_parent;
	ThreadItem[]		m_items;
	char[][]			m_current_tags;
	int					m_no;
	RegExp				m_tag_pattern1, m_tag_pattern2;
	Menu				m_menu;
	MenuItem			m_menu_remove;
	bool				m_focus_flag, m_tag_focus_flag;
	
	BrowserTabPage delegate(char[] address)	m_add_page_func;
	
	void disposeItems()
	{
		foreach (inout item; m_items) {
			item.dispose();
			delete item;
			item = null;
		}
		delete m_items;
		m_items = null;
		m_no = 0;
	}
	
	void setSortIndex(ThreadItem.SortIndex index, int direction)
	{
		ThreadItem.sortIndex = index;
		ThreadItem.sortDirection = direction;
	}
	
	ThreadItem[] tagFilter(char[] tag)
	{
		return filter(tag);
	}
	
	ThreadItem[] filter(char[] filter, ThreadItem[] ti = null)
	{
		ThreadItem[] items;
		ThreadItem[] origin;
		
		if (ti.length == 0) {
			origin = m_items;
		} else {
			origin = ti;
		}
		
		foreach (e; origin) {
			bool ok = true;
			if (filter.length == 0) {
				items ~= e;
			} else if (findKeyword(std.string.toupper(e.title), std.string.toupper(filter))) {
				items ~= e;
			}
		}
		
		return items;
	}
	
	void redrawItems()
	{
		Configuration config = getConfig();
		char[] tag_filter, pattern_filter;
		int count = m_table.getItemCount();
		int internal_count;
		ThreadCache cache = ThreadCache.getInstance();
		int[char[]] tag_table = null;
		bool null_filter;
		
		ThreadIndex[char[]] indexes;
		ThreadIndex *index;
		TableItem[] items;
		ThreadItem[] internal_items;
		
		tag_filter 		= m_tag_filter.getText();
		pattern_filter 	= m_pattern_filter.getText();
		
		if (tag_filter.length == 0 && pattern_filter.length == 0) {
			null_filter = true;
		} else {
			null_filter = false;
		}
		if (tag_filter.length != 0) {
			internal_items = tagFilter(tag_filter);
		} else {
			internal_items = m_items.sort;
		}
		if (pattern_filter.length != 0
			&& internal_items.length != 0) 
		{
			internal_items = filter(pattern_filter, internal_items);
		}
		internal_count = internal_items.length;
		internal_items = internal_items.sort;
		
		foreach (e; cache.getIndexes()) {
			indexes[e.title] = e;
		}
		
		m_table.setRedraw(false);
		scope(exit) {
			m_table.setRedraw(true);
		}
		
		if (internal_count == 0) {
			m_table.clearAll();
			m_table.setRedraw(true);
			m_table.setRedraw(false);
			m_table.setItemCount(1);
			m_table.setItemCount(0);
			return;
		}
		if (count != internal_count) {
			m_table.setItemCount(internal_count);
			count = m_table.getItemCount();
		} else {
			m_table.setItemCount(count);
		}
		
		items = m_table.getItems();
		
		Color stripe 	= new Color(null, 0xf0, 0xf0, 0xff);
		
		scope(exit) stripe.dispose();
		
		uint pt, ct;
		for (int i = 0; i < count; ++i) {
			TableItem 	table_item 		= items[i];
			ThreadItem	internal_item	= internal_items[i];
			
			index = (internal_item.title in indexes);
			
			table_item.clear();
			
			if (index != null) {
				ThreadIndex ti = *index;
				if (ti.lastPostTime != internal_item.lastModifiedTimeNum) {
					table_item.setImage(0, getIcon("page_red"));
					table_item.setForeground(1, 
						Display.getCurrent.getSystemColor(DWT.COLOR_DARK_BLUE));
				} else {
					table_item.setImage(0, getIcon("page"));
				}
				table_item.setText(3, std.string.toString(ti.count));
			} else {
				table_item.setImage(0, getIcon("page_green"));
				table_item.setText(3, internal_item.records);
			}
			table_item.setText(0, internal_item.no);
			table_item.setText(1, internal_item.title);
			table_item.setText(2, internal_item.threadSize);
			table_item.setText(4, internal_item.lastModifiedTime);
			table_item.setData(internal_item);
			
			if (i % 2 == 0) {
				table_item.setBackground = stripe;
			}
			
			if (null_filter) {
				// tag tract
/*
				foreach (e; m_tag_pattern1.match(internal_item.title)) {
					if (e.length > 2) {
						char[] tag = std.string.strip(e);
						if ((tag in tag_table) == null) {
							char[] alloc_tag = tag.dup;
							tag_table[alloc_tag] = 1;
							config.addTag(alloc_tag);
						}
					}
				}
*/
				foreach (e; getTags(internal_item.title)) {
					char[] tag = std.string.strip(e);
					if ((tag in tag_table) == null) {
						char[] alloc_tag = tag.dup;
						tag_table[alloc_tag] = 1;
						config.addTag(alloc_tag);
					}
				}
			}
		}
		if (null_filter) {
			m_current_tags = null;
			m_current_tags ~= "";
			m_current_tags ~= tag_table.keys.sort;
			if (m_current_tags.length != 0) {
				char[][] tags = config.getSortedTags(m_current_tags);
				m_tag_filter.setItems(tags);
			}
		}
	}
	
	void sortItems()
	{
		if (m_table.isDisposed) return;
		
		Object current_selection = null;
		Item[] items 	= m_table.getSelection();
		
		if (items.length != 0) {
			current_selection = items[0].getData();
		}
		
		redrawItems();
		
		m_table.deselectAll();
		if (current_selection !is null) {
			foreach (item; m_table.getItems) {
				if (current_selection is item.getData()) {
					m_table.setSelection(m_table.indexOf(item));
					break;
				}
			}
		}
	}
	
	void fileterItems()
	{
		
	}
	
	// カラムソート
	void columnSelectionHandler(SelectionEvent e)
	{
		TableColumn old_selection 		= m_table.getSortColumn;
		TableColumn current_selection	= cast(TableColumn)e.cData;
		int direction;
		ThreadItem.SortIndex index;
		
		if (old_selection is current_selection) {
			direction = m_table.getSortDirection == DWT.UP ? DWT.DOWN: DWT.UP;
		} else {
			direction = DWT.DOWN;
		}
		index = (cast(SortIndexWrapper)current_selection.getData()).getData;
		
		m_table.setSortColumn(current_selection);
		m_table.setSortDirection(direction);
		
		setSortIndex(index, direction);
		sortItems();
	}
	
	int getNo()
	{
		return ++m_no;
	}
	
	void addPage(char[] address)
	{
		if (m_add_page_func != null) {
			m_add_page_func(address);
		}
	}
	
	ThreadItem getCurrentSelectionItem()
	{
		TableItem[]	table_item 	= m_table.getSelection();
		if (table_item.length == 0) {
			return null;
		}
		
		return cast(ThreadItem)table_item[0].getData();
	}
	
	
	void doubleClickHandler(MouseEvent event)
	{
		ThreadItem internal_item = getCurrentSelectionItem();
		if (internal_item is null) {
			return;
		}
		
		Schema schema 	= new Schema(internal_item.title);
		schema.method 	= "get";
		schema.mode 	= Schema.Mode.BACK_ROWS;
		schema.items	= getConfig.items;
		
		addPage(schema.toURI);
	}
	
	void tagModifyHandler(ModifyEvent e)
	{
		sortItems();
	}
	
	void tagSelectionHandler(SelectionEvent e)
	{
		Configuration config = getConfig();
		char[] tag = m_tag_filter.getText();
		
		if (tag.length != 0) {
			config.tagPriorityUp(tag);
			if (m_current_tags.length != 0) {
				m_tag_filter.setItems(config.getSortedTags(m_current_tags));
			}
		}
	}
	
	void menuPopHandler(MouseEvent e)
	{
		int button = e.button;
		if (e.button != 3 
			|| m_table.getSelection() is null) 
		{
			return; // clicked button is not right button
		}
		
		m_menu_remove.setEnabled(m_list_selection);
		
		m_menu.setLocation(m_table.toDisplay(e.x, e.y));
		m_menu.setVisible(true);
	}
	
	void threadOpenHandler(SelectionEvent e)
	{
		ThreadItem item = getCurrentSelectionItem();
		if (item is null) {
			return;
		}
		
		Schema schema 	= new Schema(item.title);
		schema.method 	= "get";
		schema.mode 	= Schema.Mode.BACK_ROWS;
		schema.items	= getConfig.items;
		
		addPage(schema.toURI);
	}
	
	void threadGatewayOpenHandler(SelectionEvent e)
	{
		ThreadItem item = getCurrentSelectionItem();
		if (item is null) {
			return;
		}
		
		Configuration config = getConfig();
		char[] url = std.string.format("http://%s:%s/thread.cgi/%s", 
						config.nodeName, config.nodePort, std.uri.encode(item.title));
		
		addPage(url);
	}
	
	void threadBookmarkHandler(SelectionEvent e)
	{
		ThreadItem item = getCurrentSelectionItem();
		if (item is null) {
			return;
		}
		
		Schema schema 	= new Schema(item.title);
		schema.method 	= "get";
		schema.mode 	= Schema.Mode.BACK_ROWS;
		schema.items	= getConfig.items;
		
		addFavorite(m_parent.getShell(), getConfig.favoritesDirectory, schema.toURI, item.title);
	}
	
	void titleCopyHandler(SelectionEvent e)
	{
		ThreadItem item = getCurrentSelectionItem();
		if (item is null) {
			return;
		}
		copyText(item.title);
	}
	
	void shingetsuURLCopyHandler(SelectionEvent e)
	{
		ThreadItem item = getCurrentSelectionItem();
		if (item is null) {
			return;
		}
		
		copyText(std.string.format("[[%s]]", item.title));
	}
	
	void hovenURLCopyHandler(SelectionEvent e)
	{
		ThreadItem item = getCurrentSelectionItem();
		if (item is null) {
			return;
		}
		Schema schema 	= new Schema(item.title);
		schema.method 	= "get";
		
		copyText(std.string.format(schema.toURI));
	}
	
	void gatewayURLCopyHandler(SelectionEvent e)
	{
		ThreadItem item = getCurrentSelectionItem();
		if (item is null) {
			return;
		}
		Configuration config = getConfig();
		char[] url = std.string.format("http://%s:%s/thread.cgi/%s", 
						config.nodeName, config.nodePort, std.uri.encode(item.title));
		
		copyText(url);
	}
	
	class RemovedFinalizer: Runnable
	{
		void run()
		{
			sleep(1);
			if (m_list_selection) {
				registeredThreads(null);
			} else {
				recentThreads(null);
			}
		}
	}
	
	void threadRemoveHandler(SelectionEvent e)
	{
		ThreadItem item = getCurrentSelectionItem();
		if (item is null) {
			return;
		}
		Configuration config = getConfig();
		char[] url = std.string.format("http://%s:%s/admin.cgi?cmd=fdel&file=%s", 
						config.nodeName, config.nodePort, encodeThreadName(item.title));
		
		BorwserChildWindow child = new BorwserChildWindow;
		child.setOneTimeEnabled(true);
		child.setFinalizer(new RemovedFinalizer);
		child.getBrowser.setUrl(url);
		child.run();
	}
	
	
	void tagFocusOutHandler(FocusEvent e)
	{
		m_tag_focus_flag = false;
	}
	
	void tagClickHandler(MouseEvent e)
	{
		if (! m_tag_focus_flag) {
			m_tag_filter.parent.setFocus();
			m_tag_filter.text.setFocus();
			m_tag_focus_flag = true;
		}
	}
	
	void filterFocusOutHandler(FocusEvent e)
	{
		m_focus_flag = false;
	}
	
	void filterClickHandler(MouseEvent e)
	{
		if (! m_focus_flag) {
			m_pattern_filter.selectAll();
			m_focus_flag = true;
		}
	}
	
public:
	this(Composite parent, int style)
	{
		m_parent 		= new Composite(parent, DWT.NONE);
		m_style			= style;
//		m_tag_pattern1	= new RegExp("\\[[^\\]]+\\]", "g");
		m_tag_pattern2	= new RegExp("【.+】");
		createWidgets();
		
		getConfig.addTag("");
	}
	
	void createWidgets()
	{
		m_parent.setLayout(new GridLayout(4, false));
		
		with (new CLabel(m_parent, DWT.FLAT)) {
			Image		ticon 	= getIcon("tag_blue");
			
			ticon.transparentPixel = 0xffffff;	// white
			ticon.setBackground(getBackground());
			
			setImage(ticon);
			setLayoutData(new GridData(GridData.BEGINNING));
			setToolTipText("タグ");
		}
		m_tag_filter	= new CCombo(m_parent, DWT.DROP_DOWN|DWT.FLAT);
		m_tag_filter.setLayoutData(new GridData(100, -1));
		m_tag_filter.handleModify(null, &tagModifyHandler);
		m_tag_filter.handleSelection(null, &tagSelectionHandler);
		m_tag_filter.handleFocusOut(null, &tagFocusOutHandler);
		m_tag_filter.handleMouseDown(null, &tagClickHandler);
		m_tag_filter.setVisibleItemCount(50);
		
		with(new CLabel(m_parent, DWT.FLAT)) {
			Image		ficon 	= getIcon("shape_move_forwards");
			
			ficon.transparentPixel = 0xffffff; // white
			ficon.setBackground(getBackground());
			setImage(ficon);
			setLayoutData(new GridData(GridData.BEGINNING));
			setToolTipText("フィルター");
		}
		m_pattern_filter	= new Text(m_parent, DWT.SINGLE|DWT.FLAT);
		m_pattern_filter.setLayoutData(new GridData(GridData.HORIZONTAL_ALIGN_FILL));
		m_pattern_filter.handleModify(null, &tagModifyHandler);
		m_pattern_filter.handleFocusOut(null, &filterFocusOutHandler);
		m_pattern_filter.handleMouseDown(null, &filterClickHandler);
		
		GridData list_grid = new GridData(GridData.FILL_BOTH);
		list_grid.horizontalSpan = 4;
		
		m_table = new Table(m_parent, m_style);
		m_table.setLayoutData(list_grid);
		
		TableColumn no_col, title_col, size_col, energy_col, record_col, last_col;
		
		with (no_col =  new TableColumn(m_table, DWT.LEFT)) {
			setText("№");
			setWidth(60);
			setData(new SortIndexWrapper(ThreadItem.SortIndex.NO));
			handleSelection(no_col, &columnSelectionHandler);
		}
		with (title_col = new TableColumn(m_table, DWT.LEFT)) {
			setText("タイトル");
			setWidth(240);
			setData(new SortIndexWrapper(ThreadItem.SortIndex.TITLE));
			handleSelection(title_col, &columnSelectionHandler);
		}
		with (size_col = new TableColumn(m_table, DWT.RIGHT)) {
			setText("ゲートウェイ保有サイズ");
			setWidth(150);
			setData(new SortIndexWrapper(ThreadItem.SortIndex.THREADSIZE));
			handleSelection(size_col, &columnSelectionHandler);
		}
		with (record_col = new TableColumn(m_table, DWT.RIGHT)) {
			setText("投稿数");
			setWidth(80);
			setData(new SortIndexWrapper(ThreadItem.SortIndex.REC));
			handleSelection(record_col, &columnSelectionHandler);
		}
		with (last_col = new TableColumn(m_table, DWT.LEFT)) {
			setText("最終更新時刻");
			setWidth(130);
			setData(new SortIndexWrapper(ThreadItem.SortIndex.MODDATE));
			handleSelection(last_col, &columnSelectionHandler);
		}
		m_table.setSortColumn(no_col);
		m_table.setSortDirection(DWT.UP);
		
		setSortIndex(ThreadItem.SortIndex.NO, DWT.UP);
		
		m_table.setHeaderVisible(true);
//		m_table.setLinesVisible(true);
		
		m_table.handleMouseDoubleClick(null, &doubleClickHandler);
		
		// menu
		m_menu			= new Menu(m_table.getShell());
		MenuItem item;
		with (item = new MenuItem(m_menu, DWT.PUSH)) {
			setText("開く");
			setImage(getIcon("page_add"));
			handleSelection(null, &threadOpenHandler);
		}
		
		cast(void)new MenuItem(m_menu, DWT.SEPARATOR);
		
		with (new MenuItem(m_menu, DWT.PUSH)) {
			setText("ゲートウェイを開く");
			setImage(getIcon("monitor"));
			handleSelection(null, &threadGatewayOpenHandler);
		}
		
		cast(void)new MenuItem(m_menu, DWT.SEPARATOR);
		
		with (new MenuItem(m_menu, DWT.PUSH)) {
			setText("ブックマークへ追加");
			setImage(getIcon("star"));
			handleSelection(null, &threadBookmarkHandler);
		}
		
		cast(void)new MenuItem(m_menu, DWT.SEPARATOR);
		
		with (new MenuItem(m_menu, DWT.PUSH)) {
			setText("タイトルをコピー");
			handleSelection(null, &titleCopyHandler);
		}
		with (new MenuItem(m_menu, DWT.PUSH)) {
			setText("URLを新月形式でコピー");
			handleSelection(null, &shingetsuURLCopyHandler);
		}
		with (new MenuItem(m_menu, DWT.PUSH)) {
			setText("URLをHoven形式でコピー");
			handleSelection(null, &hovenURLCopyHandler);
		}
		with (new MenuItem(m_menu, DWT.PUSH)) {
			setText("ゲートウェイのURLをコピー");
			handleSelection(null, &gatewayURLCopyHandler);
		}
		
		cast(void)new MenuItem(m_menu, DWT.SEPARATOR);
		
		with (m_menu_remove = new MenuItem(m_menu, DWT.PUSH)) {
			setText("削除");
			setImage(getIcon("delete"));
			handleSelection(null, &threadRemoveHandler);
		}
		
		m_table.handleMouseUp(null, &menuPopHandler);
	}
	
	void setAddPageHandler(BrowserTabPage delegate(char[]) func)
	{
		m_add_page_func = func;
	}
	
	void clear()
	{
		m_table.clearAll();
		m_table.setItemCount(1);	// Bad
		m_table.setItemCount(0);	// Know How
	}
	
	void recentThreads(char[] filter_regexp)
	{
		class RecentThread: Runnable {
			ThreadTable parent;
			this(ThreadTable parent) { this.parent = parent; }
			void run()
			{
				GatewayIO gateway = new GatewayIO(new InternetAddress(getConfig.nodeName, getConfig.nodePort));
				
				parent.disposeItems();
				gateway.getRecentThreads(null, &parent.addThreadItem);
				parent.redrawItems();
			}
		}
		try {
			BusyIndicator.showWhile(null, new RecentThread(this));
		} catch (Exception e) {
			MessageBox.showMessage(e.toString);
		}
		
		m_list_selection = false;
	}
	
	void registeredThreads(char[] filter_regexp)
	{
		class RegisteredThread: Runnable {
			ThreadTable parent;
			this(ThreadTable parent) { this.parent = parent; }
			void run()
			{
				GatewayIO gateway = new GatewayIO(new InternetAddress(getConfig.nodeName, getConfig.nodePort));
				
				parent.disposeItems();
				gateway.getRegisteredThreads(null, &parent.addThreadItem);
				parent.redrawItems();
			}
		}
		try {
			BusyIndicator.showWhile(null, new RegisteredThread(this));
		} catch (Exception e) {
			MessageBox.showMessage(e.toString);
		}
		
		m_list_selection = true;
	}
	
	void addThreadItem(ThreadResponseEvent event)
	{
		ThreadItem item = new ThreadItem(getNo(), event);
		m_items ~= item;
	}
}

// サイドバー

class TreeItemData
{
	bool isdir;
	InternetShortcut shortcut;
	
	this(bool isdir = true)
	{
		this.isdir = isdir;
	}
	
	this(InternetShortcut shortcut)
	{
		this.shortcut = shortcut;
	}
}

class InternetShortcut
{
	char[]	url;
	char[]	title;
	char[]	path;
	
	this(){}
	
	this(char[] title, char[] url)
	{
		this.title = title;
		this.url = url;
	}
	
	this(char[] shortcut)
	{
		path	= pes.path.absPath(shortcut);
		title 	= pathToName(shortcut);
		url		= readInternetShortcut(path);
	}
	
	static char[] pathToName(char[] path)
	{
		return std.path.getName(std.path.getBaseName(path));
	}
	
	char[] getShortcutName()
	{
		return std.string.format("%s.url", title);
	}
	
	void reload()
	{
		url = readInternetShortcut(path);
	}
	
	char[] readInternetShortcut(char[] inifile)
	{
		IniFile ini = new IniFile(inifile);
		
		return ini.readString("InternetShortcut", "URL");
	}
	
	void writeInternetShortcut(char[] inifile)
	{
		IniFile ini = new IniFile(inifile);
		
		ini.writeString("InternetShortcut", "URL", url);
	}
	
	override int opCmp(Object o)
	{
		InternetShortcut rhs = cast(InternetShortcut)o;
		int cmp = std.string.icmp(rhs.title, title);
		
		return cmp;
	}
}


static void reloadFavorites(void delegate() func = null)
{
	static void delegate() s_func = null;
	
	if (func) {
		s_func = func;
	} else if (s_func != null) {
		s_func();
	}
}


class SideBar: ControlBase
{
protected:
	static const char[] SEARCH_PAGE = "search.html";

protected:
	CTabFolder	m_tab_folder;
	Tree		m_tree;
	Composite	m_parent;
	Menu		m_menu;
	MenuItem	m_menu_open;
	
	BrowserTabPage delegate(char[] address) 	m_add_page_func;
	
	void goShortcut(MouseEvent e)
	{
		TreeItem[] item = m_tree.getSelection();
		if (item.length == 0) {
			return;
		}
		TreeItemData data = cast(TreeItemData)item[0].getData();
		if (data is null) {
			return;
		}
		if (! data.isdir) {
			data.shortcut.reload();
			addPage(data.shortcut.url);
		}
	}
	
	class NewWindowListener: OpenWindowListener
	{
		void open(WindowEvent e)
		{
			BrowserTabPage page = addPage(CBrowser.ABOUT_BLANK);
			e.browser = page.getBrowser();
		}
	}
	
	void organizeBookmarkHandler(SelectionEvent e)
	{
		organizeFavorite(m_parent.getShell(), getConfig.favoritesDirectory);
	}
	
	void organizeIEFavoriteHandler(SelectionEvent e)
	{
		organizeFavorite(m_parent.getShell(), getSystemFavoriteDirectory);
	}
	
	void menuPopHandler(MouseEvent e)
	{
		if (e.button != 3) {
			return;
		}
		TreeItem[] item = m_tree.getSelection();
		if (item.length == 0) {
			return;
		}
		TreeItemData data = cast(TreeItemData)item[0].getData();
		if (data is null) {
			return;
		}
		
		m_menu_open.setEnabled(!data.isdir);
		
		m_menu.setLocation(m_tree.toDisplay(e.x, e.y));
		m_menu.setVisible(true);
	}
	
	void reloadBookmarkHandler(SelectionEvent e)
	{
		reloadFavorites();
	}
	
	void openHandler(SelectionEvent e)
	{
		TreeItem[] item = m_tree.getSelection();
		if (item.length == 0) {
			return;
		}
		TreeItemData data = cast(TreeItemData)item[0].getData();
		if (data is null) {
			return;
		}
		if (! data.isdir) {
			data.shortcut.reload();
			addPage(data.shortcut.url);
		}
	}
	
	void createWidgets()
	{
		with (m_tab_folder = new CTabFolder(m_parent, DWT.TOP | DWT.FLAT))
		{
			Color[3] 	gradient_colors;
			int[2] 		percent_array;
			
			percent_array[0] = 50;
			percent_array[1] = 100;
			
			gradient_colors[0] = Display.getDefault.getSystemColor(DWT.COLOR_TITLE_BACKGROUND);
			gradient_colors[1] = Display.getDefault.getSystemColor(DWT.COLOR_TITLE_BACKGROUND_GRADIENT);
			gradient_colors[2] = null;
			
			setTabHeight(22);
			setSelection(0);
			setSimple(false);
			setBackground(Display.getDefault.getSystemColor(DWT.COLOR_LIST_BACKGROUND));
			setForeground(Display.getDefault.getSystemColor(DWT.COLOR_LIST_FOREGROUND));
			setSelectionForeground(Display.getDefault.getSystemColor(DWT.COLOR_TITLE_FOREGROUND));
			setSelectionBackground(gradient_colors, percent_array, true);
		}
		
		CTabItem favorite;
		with (favorite = new CTabItem(m_tab_folder, DWT.NONE)) {
			setImage(getIcon("star"));
			setText("ブックマーク");
		}
		CTabItem history;
//		with (history = new CTabItem(m_tab_folder, DWT.NONE)) {
//			setImage(getIcon("clock"));
//			setText("履歴");
//		}
//		CTabItem search;
//		with (search = new CTabItem(m_tab_folder, DWT.NONE)) {
//			setImage(getIcon("group"));
//			setText("カスタム");
//		}
		
		// お気に入り
		Composite favpage = new Composite(m_tab_folder, DWT.NONE);
		favpage.setLayout(new GridLayout(1, false));
		
		ToolBar bar = new ToolBar(favpage, DWT.FLAT);
		bar.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));
		
//		with (new ToolItem(bar, DWT.PUSH)) {
//			setText("更新");
//			handleSelection(null, &reloadBookmarkHandler);
//		}
		with (new ToolItem(bar, DWT.PUSH)) {
			setText("ブックマーク");
//			setImage(getIcon("folder"));
			handleSelection(null, &organizeBookmarkHandler);
		}
		with (new ToolItem(bar, DWT.PUSH)) {
			setText("IEのお気に入り");
//			setImage(getIcon("folder_star"));
			handleSelection(null, &organizeIEFavoriteHandler);
		}
		
		m_tree = new Tree(favpage, DWT.FLAT);
		m_tree.setLayoutData(new GridData(GridData.FILL_BOTH));
		m_tree.handleMouseDoubleClick(null, &goShortcut);
		m_tree.handleMouseUp(null, &menuPopHandler);
		
		reloadFavorites();
		favorite.setControl(favpage);
		
		// 履歴
		//history.setControl(new Composite(m_tab_folder, DWT.NONE));
		
		// 検索
//		CBrowser browser = new CBrowser(m_tab_folder, DWT.NONE);
//		search.setControl(browser);
//		browser.addOpenWindowListener(new NewWindowListener);
//		browser.setUrl(absPath(getConfig.userPage));
		
		// menu
		m_menu = new Menu(m_parent.getShell());
		with (m_menu_open = new MenuItem(m_menu, DWT.PUSH)) {
			setText("開く");
			setImage(getIcon("page_add"));
			handleSelection(null, &openHandler);
		}
		
		.reloadFavorites(&this.reloadFavorites);
	}
	
	void reloadFavorites()
	{
		TreeItem root;
		
		m_tree.setRedraw(false);
		scope(exit) m_tree.setRedraw(true);
		
		m_tree.removeAll();
		
		readBookmarkTree(m_tree, getConfig.favoritesDirectory);
		
//		root.setText("ブックマーク");
//		root.setImage(getIcon("folder"));
//		readSystemFavoriteTree(root, getConfig.favoritesDirectory);
		expandTree(m_tree);
		
		root = new TreeItem(m_tree, DWT.NONE);
		root.setText("IEのお気に入り");
		root.setImage(getIcon("folder_star"));
		readSystemFavoriteTree(root, getSystemFavoriteDirectory);
	}
	
	void setItemDetails(TreeItem item, DirEntry *dirent)
	{
		if (dirent.isdir) {
			item.setData(new TreeItemData());
			item.setText(std.path.getBaseName(dirent.name));
		} else {
			InternetShortcut  fav = new InternetShortcut(dirent.name);
			item.setData(new TreeItemData(fav));
			item.setText(fav.title);
		}
	}
	
	void expandTree(Tree root)
	{
		void expandItem(TreeItem item)
		{
			item.setExpanded(true);
			foreach (e; item.getItems) {
				expandItem(e);
			}
		}
		
		foreach (item; root.getItems) {
			if (item) {
				expandItem(item);
			}
		}
	}
	
	void readBookmarkTree(Tree root, char[] directory)
	{
		TreeItem current_item;
		
		DirEntry[] listdirent(char[] path)
		{
			DirEntry[] entries;
			bool callback(DirEntry *de)
			{
				entries ~= *de;
				return true;
			}
			listdir(path, &callback);
			
			return entries;
		}
		
		void setFileItems(char[] dir)
		{
			DirEntry[] entries = listdirent(dir);
			foreach (de; entries) {
				if (de.isdir) {
					TreeItem prev_item = current_item;
					current_item = new TreeItem(current_item, DWT.NONE);
					current_item.setImage(getIcon("folder"));
					setItemDetails(current_item, &de);
					
					setFileItems(de.name);
					current_item = prev_item;
				}
			}
			foreach (de; entries) {
				if (!de.isdir) {
					if (std.string.icmp(getExt(de.name), "url") == 0) {
						TreeItem item = new TreeItem(current_item, DWT.NONE);
						Image icon = getIcon("page_url");
//						icon.transparentPixel = 0xffffff; // white
//						icon.setBackground(item.getBackground());
						item.setImage(icon);
						setItemDetails(item, &de);
					}
				}
			}
		}
		
		DirEntry[] entries = listdirent(directory);
		current_item = null;
		foreach (de; entries) {
			if (de.isdir) {
				current_item = new TreeItem(root, DWT.NONE);
				current_item.setImage(getIcon("folder"));
				setItemDetails(current_item, &de);
				
				setFileItems(de.name);
			}
		}
		foreach (de; entries) {
			if (!de.isdir) {
				if (std.string.icmp(getExt(de.name), "url") == 0) {
					TreeItem item = new TreeItem(root, DWT.NONE);
					item.setImage(getIcon("page_url"));
					setItemDetails(item, &de);
				}
			}
		}
	}
	
	void readSystemFavoriteTree(TreeItem root, char[] directory)
	{
		TreeItem current_item;
		DirEntry[] entries;
		
		current_item = root;
		
		DirEntry[] listdirent(char[] path)
		{
			DirEntry[] entries;
			bool callback(DirEntry *de)
			{
				entries ~= *de;
				return true;
			}
			listdir(path, &callback);
			
			return entries;
		}
		
		void setFileItems(char[] dir)
		{
			DirEntry[] entries = listdirent(dir);
			foreach (de; entries) {
				if (de.isdir) {
					TreeItem prev_item = current_item;
					current_item = new TreeItem(current_item, DWT.NONE);
					current_item.setImage(getIcon("folder"));
					setItemDetails(current_item, &de);
					
					setFileItems(de.name);
					current_item = prev_item;
				}
			}
			foreach (de; entries) {
				if (!de.isdir) {
					if (std.string.icmp(getExt(de.name), "url") == 0) {
						TreeItem item = new TreeItem(current_item, DWT.NONE);
						item.setImage(getIcon("page_url"));
						setItemDetails(item, &de);
					}
				}
			}
		}
		
		setFileItems(directory);
	}
	
public:
	this(Composite parent, int style)
	{
		m_parent = new Composite(parent, DWT.NONE);
		m_parent.setLayout(new FillLayout);
		createWidgets();
	}
	
	void setAddPageHandler(BrowserTabPage delegate(char[]) func)
	{
		m_add_page_func = func;
	}
	
	BrowserTabPage addPage(char[] address)
	{
		if (m_add_page_func == null) {
			return null;
		}
		
		return m_add_page_func(address);
	}
}


// ブラウザ
class SakuBrowser: ControlBase
{
protected:
	ToolBar				m_topbar;		// トップバー
	Tree				m_tree;			// ディレクトリツリー
	TabBrowser 			m_browser;		// タブブラウザ
	SideBar				m_sidebar;		// サイドバー
	Composite			m_parent;
	ThreadTable			m_table;
	int					m_tree_percent;
	SashForm			m_horizontal_sash;
	SashForm			m_vertical_sash;
	
	void fileDoubleClickHandler(MouseEvent e)
	{
		Tree tree = cast(Tree)e.cData;
		TreeItem[] selection = tree.getSelection;
		DirEntry de;
		
		if (selection.length != 1) {
			return;
		}
		
		if (!getItemDetails(selection[0], de)) {
			loadFiles();
		
		} else if (de.isfile) {
			m_browser.changePage(std.path.join(getcwd(), de.name));
		}
	}
	
	void recentClickHandler(SelectionEvent e)
	{
		m_table.recentThreads(null);
	}
	
	void listClickHandler(SelectionEvent e)
	{
		m_table.registeredThreads(null);
	}
	
public:
	this(Composite parent)
	{
		super();
		
//		initializeSession();
		m_parent= parent;
	}
	
	void finalize()
	{
		m_browser.finalize();
	}
	
	void createWidgets()
	{
		int weights[2];
		
		m_horizontal_sash = new SashForm(m_parent, DWT.HORIZONTAL|DWT.FLAT);
		m_horizontal_sash.setBackground = m_parent.getDisplay.getSystemColor(DWT.COLOR_WIDGET_HIGHLIGHT_SHADOW);
//		m_horizontal_sash.setLayoutData(new GridData(GridData.FILL_BOTH));
		
		
		m_sidebar = new SideBar(m_horizontal_sash, DWT.NONE);
		
		m_vertical_sash = new SashForm(m_horizontal_sash, DWT.VERTICAL|DWT.FLAT);
		m_vertical_sash.setBackground = m_parent.getDisplay.getSystemColor(DWT.COLOR_WIDGET_HIGHLIGHT_SHADOW);
		
		m_table = new ThreadTable(m_vertical_sash,
			DWT.VIRTUAL|
			DWT.FULL_SELECTION|DWT.H_SCROLL| DWT.V_SCROLL|DWT.NO_BACKGROUND|DWT.BORDER);
		
		m_browser 	= new TabBrowser(m_vertical_sash);
		m_browser.handleRecentClick(null, &recentClickHandler);
		m_browser.handleListClick(null, &listClickHandler);
		
		// register delegates
		m_table.setAddPageHandler(&m_browser.changePage);
		m_sidebar.setAddPageHandler(&m_browser.changePage);
		
		// set default page size
	}
	
	void destroyWidgets()
	{
		
	}
	
	void loadFiles()
	{
		TreeItem current_item;
		
		bool setFileItems(DirEntry *de)
		{
			if (de.isdir) {
				TreeItem prev_item = current_item;
				current_item = new TreeItem(current_item, DWT.NONE);
				current_item.setText(getBaseName(de.name));
				current_item.setImage(getIcon("folder"));
				
				setItemDetails(current_item, de);
				
				listdir(de.name, &setFileItems);
				current_item = prev_item;
			} else {
				TreeItem item = new TreeItem(current_item, DWT.NONE);
				item.setText(getBaseName(de.name));
				item.setImage(getContentTypeIcon(getExt(de.name)));
				setItemDetails(item, de);
			}
			
			return true;
		}
		
		m_tree.removeAll();
		
		// 新着
		current_item = new TreeItem(m_tree, DWT.NONE);
		current_item.setText("新着投稿");
		current_item.setImage(getIcon("folder"));
		readUpdatedThreads(current_item);
		
		// フィルタ
		current_item = new TreeItem(m_tree, DWT.NONE);
		current_item.setText("フィルタ");
		current_item.setImage(getIcon("folder"));
		readThreadFilters(current_item);
		
		// 一覧
		current_item = new TreeItem(m_tree, DWT.NONE);
		current_item.setText("スレッド一覧");
		current_item.setImage(getIcon("folder"));
		readThreads(current_item);
		
		// 最近書き込んだスレッド
		current_item = new TreeItem(m_tree, DWT.NONE);
		current_item.setText("最近書き込んだスレッド");
		current_item.setImage(getIcon("folder"));
		readThreads(current_item);
		
		// 最近読んだスレッド
		current_item = new TreeItem(m_tree, DWT.NONE);
		current_item.setText("最近読んだスレッド");
		current_item.setImage(getIcon("folder"));
		readThreads(current_item);
		
	}
	
	void readUpdatedThreads(TreeItem current_item)
	{
		dummyitems(current_item);
	}
	
	void readThreadFilters(TreeItem current_item)
	{
		dummyitems(current_item);
		
	}
	void readThreads(TreeItem current_item)
	{
		dummyitems(current_item);
		
	}
	
	void dummyitems(TreeItem current_item)
	{
		TreeItem item;
		for (int i = 0; i < 5; ++i) {
			item = new TreeItem(current_item, DWT.NONE);
			item.setText("スレッド" ~ std.string.toString(i));
			if (i % 2) {
				item.setImage(getIcon("page"));
			} else {
				item.setImage(getIcon("page_red"));
			}
		}
	}
	
	static void setItemDetails(TreeItem item, DirEntry *de)
	{
		if (de == null) {
			item.setData("DirEntry", cast(Object)null);
		} else {
			item.setData("DirEntry", new WrapperObject!(DirEntry)(*de));
		}
	}
	
	static bool getItemDetails(TreeItem item, out DirEntry de)
	{
		Object data = item.getData("DirEntry");
		if (data is null) {
			return false;
		}
		
		de = (cast(WrapperObject!(DirEntry))data).getData();
		
		return true;
	}
}

