/** 設定(config.db)			*/

module config;

private import std.socket;
private import std.path;
private import pes.path;
private import pes.data.sqliteserializer;
private import pes.variant;

/** バージョン		*/
const char[] MAIN_TITLE 	= "Hoven a";
const float  MAIN_VERSION	= 0.14;

/** 表示用バージョン文字列取得	*/
char[] getVersionString(float ver = MAIN_VERSION)
{
	return std.string.format("%s %.2f", MAIN_TITLE, ver);
}


/**
 * 設定情報のテーブル名
 * 
 * 設定項目を増やす場合などに変更する
 * 変更する場合は、全バージョンの設定値をコピーする処理の追加が必要
 */
char[] getConfigTableName()
{
//	return std.string.replace(std.string.replace(getVersionString, " ", "_"), ".", "_");

	// 0.1固定
	return std.string.replace(std.string.replace(getVersionString(0.1), " ", "_"), ".", "_");
}
/* 設定		*/
class Configuration
{
	char[]				nodeName;
	ushort				nodePort;
	InternetAddress 	gateway;
	
	uint				items;
	
	bool				controlDaemon;
	char[]				cacheFile;
	char[]				attachmentsDirectory;
	char[]				templateDirectory;
	char[]				theme;
	char[]				_temporaryDirectory;
	char[]				userPage;
	char[]				_favoritesDirectory;
	char[]				homeURL;
	char[]				sakuPath;
	
	TagItem[char[]]		m_tag_items;
	
	TagItem[] getTagItems()
	{
		return m_tag_items.values;
	}
	
	char[] saku()
	{
		return absPath(sakuPath);
	}
	
	void addTag(char[] tag)
	{
		TagItem *pitem = (tag in m_tag_items);
		if (pitem == null) {
			TagItem item = new TagItem(tag);
			if (tag.length == 0) {
				item.priority = item.priority.max;
			}
			m_tag_items[tag] = item;
		}
	}
	
	char[][] getSortedTags(char[][] tags)
	{
		TagItem[] items;
		char[][] sorted_tags;
		
		foreach (e; tags) {
			items ~= m_tag_items[e];
		}
		foreach (e; items.sort) {
			sorted_tags ~= e.tag;
		}
		
		return sorted_tags;
	}
	
	void tagPriorityUp(char[] tag)
	{
		TagItem *pitem = (tag in m_tag_items);
		if (pitem != null) {
			(*pitem).priorityUp();
		}
	}
	
	char[] gatewayString()
	{
//		return std.string.format("http://%s/", gateway.toString);
		return std.string.format("http://%s:%s/", nodeName, nodePort);
	}
	
	char[] temporaryDirectory()
	{
		return absPath(_temporaryDirectory);
	}
	
	char[] favoritesDirectory()
	{
		return absPath(_favoritesDirectory);
	}
	
	void temporaryDirectory(char[] path)
	{
		_temporaryDirectory = path;
	}
	
	char[] xslTemplate()
	{
		return std.path.join(absPath(themeDirectory), "style.xsl");
	}
	
	static char[] xslTemplate(char[] theme_directory, char[] method)
	{
		return std.path.join(absPath(theme_directory), method ~ ".xsl");
	}
	
	char[] themeDirectory()
	{
		return std.path.join(absPath(templateDirectory), theme);
	}
	
	void serialize()
	{
		
		SQLiteSerializer serializer = new SQLiteSerializer("config.db", getConfigTableName);
		Configuration config = new Configuration;
		
		VArray array = VArray.make(
			nodeName,
			cast(long)nodePort,
			cast(long)(controlDaemon ? 1:0),
			theme,
			cast(long)items,
			homeURL,
			sakuPath
		);
		int onoff = controlDaemon;
		
		serializer.serialize(array);
	}
	
	void deserialize()
	{
		SQLiteSerializer deserializer = new SQLiteSerializer("config.db", getConfigTableName);
		VArray  array;
		try {
			array = deserializer.deserialize();
		} catch (Exception ) {
			return;
		}
		if (array.toLong == 7) {
			nodeName		= array[0].toString;
			nodePort		= cast(ushort)array[1].toLong;
			controlDaemon	= array[2].toLong == 1 ? true:false;
			theme			= array[3].toString;
			items			= cast(uint)array[4].toLong;
			homeURL			= array[5].toString;
			sakuPath		= array[6].toString;
		}
		
		int onoff = controlDaemon;
	}
}

private Configuration g_config;

static this()
{
	g_config 						= new Configuration;
	g_config.nodeName				= "localhost";
	g_config.nodePort				= 8000;
	g_config.controlDaemon			= true;
	g_config.cacheFile				= "cache.db";
	g_config.attachmentsDirectory	= "attach";
	g_config.templateDirectory		= "template" ~ std.path.sep ~ "theme";
	g_config.userPage				= "template" ~ std.path.sep ~ "user" ~std.path.sep ~ "index.html";
	g_config.theme					= "terminal";
	g_config.temporaryDirectory		= "temporary";
	g_config.items					= 30;
	g_config._favoritesDirectory	= "Bookmark";
	g_config.homeURL				= "http://www.google.co.jp/";
	g_config.sakuPath				= "winsaku" ~ std.path.sep ~ "tksaku.exe";
	
	// deserialize
	g_config.deserialize();
	
	try {
		g_config.gateway 				= new InternetAddress(g_config.nodeName, g_config.nodePort);
	} catch (Exception){}
}

static ~this()
{
	g_config.serialize();
}

Configuration getConfig()
{
	return g_config;
}




/* 優先度付きタグアイテム(検索リスト) */
class TagItem
{
	char[]	tag;
	int	priority;
	
	this(char[] tag)
	{
		this.tag = tag;
		priority = 0;
	}
	
	override int opCmp(Object o)
	{
		TagItem rhs = cast(TagItem)o;
		int cmp = rhs.priority - priority;
		if (cmp == 0) {
			cmp = std.string.icmp(rhs.tag, tag);
		}
		
		return cmp;
	}
	
	void priorityUp()
	{
		if (priority < priority.max) {
			++priority;
		}
	}
	
	void priorityDown()
	{
		if (priority <= 0) {
			--priority;
		} else {
			priority = 0;
		}
	}
}
