module daemonmanager;
/**
 * デーモンマネージャ
 * 
 * Win32 exe の制御
 */
 
private import std.c.string;
private import std.utf;
private import std.c.windows.windows;
private import std.windows.syserror;

const DWORD STARTF_USESHOWWINDOW	= 0x00000001;
/*
#define STARTF_USESIZE          0x00000002
#define STARTF_USEPOSITION      0x00000004
#define STARTF_USECOUNTCHARS    0x00000008
#define STARTF_USEFILLATTRIBUTE 0x00000010
#define STARTF_RUNFULLSCREEN    0x00000020
#define STARTF_FORCEONFEEDBACK  0x00000040
#define STARTF_FORCEOFFFEEDBACK 0x00000080
#define STARTF_USESTDHANDLES    0x00000100
*/
const DWORD WAIT_TIMEOUT			= 258;

struct PROCESS_INFORMATION {
    HANDLE hProcess;
    HANDLE hThread;
    DWORD dwProcessId;
    DWORD dwThreadId;
}

struct STARTUPINFOW { // _STARTUPINFOW
    DWORD   cb;
    LPWSTR  lpReserved;
    LPWSTR  lpDesktop;
    LPWSTR  lpTitle;
    DWORD   dwX;
    DWORD   dwY;
    DWORD   dwXSize;
    DWORD   dwYSize;
    DWORD   dwXCountChars;
    DWORD   dwYCountChars;
    DWORD   dwFillAttribute;
    DWORD   dwFlags;
    WORD    wShowWindow;
    WORD    cbReserved2;
    LPBYTE  lpReserved2;
    HANDLE  hStdInput;
    HANDLE  hStdOutput;
    HANDLE  hStdError;
}
alias STARTUPINFOW STARTUPINFO;

extern(Windows)
{
	void ZeroMemory(VOID *p, DWORD len)
	{
		memset(p, 0, len);
	}
	
	BOOL CreateProcessW(
		LPCWSTR lpApplicationName,
		LPWSTR lpCommandLine,
		LPSECURITY_ATTRIBUTES lpProcessAttributes,
		LPSECURITY_ATTRIBUTES lpThreadAttributes,
		BOOL bInheritHandles,
		DWORD dwCreationFlags,
		LPVOID lpEnvironment,
		LPCWSTR lpCurrentDirectory,
		STARTUPINFOW *lpStartupInfo,
		PROCESS_INFORMATION *lpProcessInformation
	);
	
	BOOL TerminateProcess(
		HANDLE hProcess,
		UINT uExitCode
	);
	
	DWORD WaitForInputIdle(
		HANDLE hProcess,
		DWORD dwMilliseconds
	);
	
	DWORD GetWindowTextA(HWND hwnd, char *ptr, uint len);
	DWORD SetWindowTextA(HWND hwnd, char *ptt);

	DWORD GetWindowThreadProcessId(HWND hwnd, void *);
	
	DWORD EnumWindows(BOOL function(HWND, LPARAM), LPARAM);
	DWORD EnumChildWindows(HWND, BOOL function(HWND, LPARAM), LPARAM);
	
	DWORD GetWindowLongA(HWND, DWORD);
	DWORD SetWindowLongA(HWND, DWORD, DWORD);
	
	BOOL EnableWindow(HWND, BOOL);
	
	DWORD GetClassNameA(HWND, LPCTSTR, DWORD);
}
alias GetWindowLongA GetWindowLong;
alias SetWindowLongA SetWindowLong;
alias GetClassNameA GetClassName;

alias CreateProcessW CreateProcess ;
alias GetWindowTextA GetWindowText;
alias SetWindowTextA SetWindowText;

class DaemonManagerException: Exception
{
	this(char[] msg)
	{
		super(msg);
	}
}

class DaemonManager
{
protected:
	PROCESS_INFORMATION	m_program;
	bool	m_active;
	char[]	m_path;
	HWND	m_handle = cast(HWND)0;
	
	void activate()
	{
		m_active = true;
	}
	
	extern (Windows) 
	{
		struct FindWindowParam {
			DWORD 	threadId;
			HWND	hWnd;
		};
		
		struct FindControlParam {
			char *szText;
			HWND hWnd;
		};
		
		struct FindClassParam {
			char *szText;
			HWND[] hWnds;
		};
		
		static BOOL findWindowProc(HWND hWnd, LPARAM lParam)
		{
			FindWindowParam *pParam = cast(FindWindowParam *)lParam;
			if (pParam.threadId == GetWindowThreadProcessId(hWnd, null)) {
				
				pParam.hWnd = hWnd;
				
				return FALSE;
			}
			
			return TRUE;
		}
		
		static BOOL findControlProc(HWND hWnd, LPARAM lParam)
		{
			FindControlParam *pParam	= cast(FindControlParam *)lParam;
			char[1024] szControlText;
			
			GetWindowText(hWnd, szControlText.ptr, szControlText.length);
			if (strcmp(pParam.szText, szControlText) == 0) {
				pParam.hWnd = hWnd;
				return FALSE;
			}
			
			return TRUE;
		}
		
		static BOOL findClassProc(HWND hWnd, LPARAM lParam)
		{
			FindClassParam *pParam	= cast(FindClassParam *)lParam;
			char[1024] szClassText;
			
			GetClassName(hWnd, szClassText.ptr, szClassText.length);
			if (strcmp(pParam.szText, szClassText) == 0) {
				pParam.hWnds ~= hWnd;
			}
			
			return TRUE;
		}
	}
	
	char[] m_argv;
	
public:
	const uint	INFINITE	= cast(uint)-1;
	
	this(char[] path)
	{
		m_path = path;
	}
	
	this(char[] path, char[] argv)
	{
		this(path);
		m_argv = argv;
	}
	
	/**
	 * 開始
	 *
	 */
	void start()
	{
		STARTUPINFO stStartupInfo;
		BOOL result;
		wchar *argv = m_argv.length ? std.utf.toUTF16z(m_argv):null;
		
		ZeroMemory(&stStartupInfo, stStartupInfo.sizeof);
		ZeroMemory(&m_program, m_program.sizeof);
		
		stStartupInfo.cb			= stStartupInfo.sizeof;
		stStartupInfo.dwFlags		= STARTF_USESHOWWINDOW ;
		stStartupInfo.wShowWindow 	= SW_HIDE;
		
		result = CreateProcess(std.utf.toUTF16z(m_path), argv, null, null, FALSE, 0,
			null, std.utf.toUTF16z(std.path.getDirName(m_path)), &stStartupInfo, &m_program);
		
		if (result == FALSE) {
			throw new DaemonManagerException(sysErrorString(GetLastError()));
		}
// tksaku
//		waitForInputIdle(INFINITE);
//		m_handle = getWindowHandle();
//		if (m_handle) {
//			setVisible(false);
//		}
		
		activate();
	}
	
	void setVisible(bool sw)
	{
		if (m_handle is null) {
			m_handle = getWindowHandle();
		}
		ShowWindow(m_handle, sw ? TRUE: FALSE);
	}
	
	bool waitForInputIdle(DWORD msec)
	{
		DWORD result = WaitForInputIdle(m_program.hProcess, msec);
		if (result == WAIT_TIMEOUT) {
			return false; // timeout
		} else if (result != 0) {
			throw new DaemonManagerException(sysErrorString(GetLastError()));
		}
		
		return true; // result == 0
	}
	
	bool waitForSingleObject(DWORD msec)
	{
		DWORD result = WaitForSingleObject(m_program.hProcess, msec);
		if (result == WAIT_TIMEOUT) {
			return false; // timeout
		} else if (result != 0) {
			throw new DaemonManagerException(sysErrorString(GetLastError()));
		}
		
		return true; // result == 0
	}
	
	/**
	 * 停止
	 *
	 */
	void kill()
	{
		DWORD result = 0;
		DWORD first_error = 0;
		
		bool has_error = false;
		
		if (this.isActive) {
			result = CloseHandle(m_program.hThread);
			if (result == FALSE) {
				if (!has_error) {
					has_error = true;
					first_error = GetLastError();
				}
			}
			result = TerminateProcess(m_program.hProcess, 0);
			if (result == FALSE) {
				if (!has_error) {
					has_error = true;
					first_error = GetLastError();
				}
			}
			result = WaitForSingleObject(m_program.hProcess, INFINITE);
			if (result != FALSE) {
				if (!has_error) {
					has_error = true;
					first_error = GetLastError();
				}
			}
			result = CloseHandle(m_program.hProcess);
			if (result == FALSE) {
				if (!has_error) {
					has_error = true;
					first_error = GetLastError();
				}
			}
			if (has_error) {
				throw new DaemonManagerException(sysErrorString(first_error));
			}
		}
	}
	
	/**
	 * 再起動
	 *
	 */
	void restart()
	{
		kill();
		start();
	}
	
	bool isActive()
	{
		return m_active;
	}
	
	HWND getHandle()
	{
		return m_handle;
	}
	
	HWND getControlHandle(HWND hWindowHandle, char[] label)
	{
		FindControlParam param;
		
		param.hWnd = cast(HWND)0;
		param.szText = std.string.toStringz(label);
		
		EnumChildWindows(hWindowHandle, &findControlProc, cast(LPARAM)&param);
		
		return param.hWnd;
	}
	
	enum WindowStyle: uint 
	{
		WS_SYSMENU          = 0x00080000L,
		WS_MINIMIZEBOX      = 0x00020000L,
		WS_MAXIMIZEBOX      = 0x00010000L,
		WS_CAPTION          = 0x00C00000L,
	};
	const uint GWL_STYLE  = cast(uint)(-16);
	
	void enableWindowStyle(uint style)
	{
		if (m_handle) {
			uint cur = GetWindowLong(m_handle, GWL_STYLE);
			SetWindowLong(m_handle, GWL_STYLE, cur | style);
		}
	}
	
	void disableWindowStyle(uint style)
	{
		if (m_handle) {
			uint cur = GetWindowLong(m_handle, GWL_STYLE);
			SetWindowLong(m_handle, GWL_STYLE, cur & ~style);
		}
	}
	
	void setControlEnabled(char[] label, bool sw)
	{
		if (m_handle) {
			HWND handle = getControlHandle(m_handle, label);
			if (handle) {
				EnableWindow(handle, sw ? TRUE:FALSE);
			}
		}
	}
	
	void setClassEnabled(char[]class_name, bool sw)
	{
		if (m_handle) {
			HWND[] handles = getClassHandles(m_handle, class_name);
			foreach (handle; handles) {
				EnableWindow(handle, sw ? TRUE:FALSE);
			}
		}
	}
	
	void setClassVisible(char[]class_name, bool sw)
	{
		if (m_handle) {
			HWND[] handles = getClassHandles(m_handle, class_name);
			foreach (handle; handles) {
				ShowWindow(handle, sw ? TRUE:FALSE);
			}
		}
	}
	
	void setClassText(char[] class_name, char[] text)
	{
		if (m_handle) {
			HWND[] handles = getClassHandles(m_handle, class_name);
			foreach (handle; handles) {
				SetWindowText(handle, std.string.toStringz(text));
			}
		}
	}
	
	HWND getWindowHandle()
	{
		FindWindowParam param;
		
		param.hWnd = cast(HWND)0;
		param.threadId = m_program.dwThreadId;
		
		EnumWindows(&findWindowProc, cast(LPARAM)&param);
		
		return param.hWnd;
	}
	
	HWND[] getClassHandles(HWND hWindowHandle, char[] class_name)
	{
		FindClassParam param;
		
		param.hWnds = null;
		param.szText = std.string.toStringz(class_name);
		
		EnumChildWindows(hWindowHandle, &findClassProc, cast(LPARAM)&param);
		
		return param.hWnds;
	}
}

debug(daemonmanager)
{
	import std.stdio;
	int main()
	{
		DaemonManager p = new DaemonManager(`calc.exe`);
		bool result;
		
		try {
			p.start();
			result = p.waitForInputIdle(-1);
			writefln("%s", result);
			Sleep(5000);
			p.restart();
			Sleep(5000);
			
			p.kill();
		} catch (Exception e) {
			writef("error: %s\n", e.toString);
		}
		
		return 0;
	}
}
