/** メインウィンドウ	*/
module main;

private import std.string;
private import std.process;
private import std.c.windows.windows;
private import std.windows.charset;
private import pes.variant;
private import pes.data.sqlite3;
private import pes.wrapper;

private import browser;
private import daemonmanager;
private import icon;
private import control;
private import config;

import dwt.all;
import std.c.windows.windows;

version(build){
	debug{
		pragma(lib, "dwtd.lib");
	}else{
		pragma(lib, "dwt.lib");
	}
	pragma(lib, "advapi32.lib");
	pragma(lib, "comctl32.lib");
	pragma(lib, "gdi32.lib");
	pragma(lib, "shell32.lib");
	pragma(lib, "comdlg32.lib");
	pragma(lib, "ole32.lib");
	pragma(lib, "uuid.lib");
	pragma(lib, "phobos.lib");
	
	pragma(lib, "user32_dwt.lib");
	pragma(lib, "imm32_dwt.lib");
	pragma(lib, "shell32_dwt.lib");
	pragma(lib, "msimg32_dwt.lib");
	pragma(lib, "gdi32_dwt.lib");
	pragma(lib, "kernel32_dwt.lib");
	pragma(lib, "usp10_dwt.lib");
	pragma(lib, "olepro32_dwt.lib");
	pragma(lib, "oleaut32_dwt.lib");
	pragma(lib, "oleacc_dwt.lib");	
}

extern(Windows)
{

	struct SHELLEXECUTEINFOA {
		DWORD cbSize;
		ULONG fMask;
		HWND hwnd;
		LPCSTR   lpVerb;
		LPCSTR   lpFile;
		LPCSTR   lpParameters;
		LPCSTR   lpDirectory;
		int nShow;
		HINSTANCE hInstApp;
		// Optional fields
		LPVOID lpIDList;
		LPCSTR   lpClass;
		HKEY hkeyClass;
		DWORD dwHotKey;
		union {
		HANDLE hIcon;
		HANDLE hMonitor;
		};
		HANDLE hProcess;
	};
	alias SHELLEXECUTEINFOA SHELLEXECUTEINFO;
	alias SHELLEXECUTEINFO* LPSHELLEXECUTEINFO;
	
	const uint SW_RESTORE = 9;
	const uint SEE_MASK_NOCLOSEPROCESS  =  0x00000040;
	const uint WAIT_TIMEOUT             = 258L;    // dderror
	
	DWORD WaitForSingleObject(
		HANDLE hHandle,
		DWORD dwMilliseconds
	);
	
	HINSTANCE ShellExecuteA(
	    HWND hwnd, 
		LPCTSTR lpVerb,
	    LPCTSTR lpFile, 
	    LPCTSTR lpParameters,
	    LPCTSTR lpDirectory,
	    INT nShowCmd
	);
	BOOL ShellExecuteExA(
		LPSHELLEXECUTEINFO lpExecInfo
	);
	BOOL PeekMessageA(MSG*, HWND, UINT, UINT, UINT);
	BOOL TranslateMesage(MSG*);
	BOOL DispatchMessageA(MSG*);
	
	DWORD GetWindowThreadProcessId(HWND hwnd, void *);
	
	DWORD EnumWindows(BOOL function(HWND, LPARAM), LPARAM);
	DWORD EnumChildWindows(HWND, BOOL function(HWND, LPARAM), LPARAM);
	
	DWORD ProcessHandleToId(HANDLE);
	
	UINT GetSystemDirectoryA(
	  LPTSTR lpBuffer,  // システムディレクトリが格納されるバッファ
	  UINT uSize        // ディレクトリバッファのサイズ
	);
}
alias GetSystemDirectoryA GetSystemDirectory;
alias DispatchMessageA DispatchMessage;
alias PeekMessageA PeekMessage;
alias ShellExecuteExA ShellExecuteEx;
alias ShellExecuteA ShellExecute;

int main()
{
	MainForm form = new MainForm;
	
	form.run();
	
	return 0;
}

class MainForm: ControlBase
{
protected:
	Shell m_shell;
	Display m_display;
	TrayItem m_tray_item;
	Menu m_tray_menu;
	SakuBrowser m_browser;
	DaemonManager m_daemon;
	
	extern (Windows) 
	{
		struct FindWindowParam {
			DWORD 	threadId;
			HWND	hWnd;
		};
		
		
		static BOOL findWindowProc(HWND hWnd, LPARAM lParam)
		{
			FindWindowParam *pParam = cast(FindWindowParam *)lParam;
			if (pParam.threadId == GetWindowThreadProcessId(hWnd, null)) {
				
				pParam.hWnd = hWnd;
				
				return FALSE;
			}
			
			return TRUE;
		}
	}
	
	HWND getWindowHandle(DWORD process_id)
	{
		FindWindowParam param;
		
		param.hWnd = cast(HWND)0;
		param.threadId = process_id;
		
		EnumWindows(&findWindowProc, cast(LPARAM)&param);
		
		return param.hWnd;
	}
	
	enum TrayItemId: uint
	{
		VISIBLE		= 0,
		EXIT		= 1
	}
	
	void serialize()
	{
		long maximized = m_shell.getMaximized;
		
		if (maximized != 0) {
			m_shell.setMaximized = 0;
		}
		
		setGUIConfig(
			VArray.make(
			maximized,
			cast(long)m_shell.getSize.x,
			cast(long)m_shell.getSize.y,
			cast(long)m_shell.getLocation.x,
			cast(long)m_shell.getLocation.y)
		);
	}
	
	void deserialize()
	{
		try {
			VArray array = getGUIConfig();
			if (array.toLong == 5) {
				m_shell.setSize(array[1].toLong, array[2].toLong);
				m_shell.setLocation(array[3].toLong, array[4].toLong);
				m_shell.setMaximized = array[0].toBool;
			}
		}  catch (SQLiteException e) {
			return; // first
		}
	}
	
	void finalize()
	{
		m_browser.finalize();
		m_tray_item.dispose();
	}
	
	void closeHandler(ShellEvent e)
	{
		bool doit;
		int style = DWT.APPLICATION_MODAL | DWT.YES | DWT.NO;
		char[] program = getVersionString();
		
		MessageBox dlg = new MessageBox (m_shell, style);
		dlg.setText(program);
		dlg.setMessage (std.string.format("%s終了しますか?", program));
		doit = dlg.open () == DWT.YES;
		
		if (doit) {
			serialize();
			finalize();
		}
		e.doit = doit;
	}
	
	void exitHandler(SelectionEvent e)
	{
		m_shell.close();
	}
	
	void visibleTkSakuHandler(SelectionEvent e)
	{
		MenuItem item = cast(MenuItem)e.widget;
		if (m_daemon) {
			m_daemon.setVisible(item.getSelection() ? true:false);
		} else {
			if (item.getSelection) {
				MessageBox.showMessage("朔の制御が無効です");
				item.setSelection(0);
			}
		}
	}
	
	void doAboutDialog(SelectionEvent e)
	{
		AboutDialog dialog = new AboutDialog();
		dialog.run();
	}
	
	void doInternetOptionDialog(SelectionEvent e)
	{
		SHELLEXECUTEINFO info;
		MSG msg;
		
		info.cbSize			= SHELLEXECUTEINFO.sizeof;
		info.fMask			= SEE_MASK_NOCLOSEPROCESS;
		info.hwnd			= m_shell.handle;
		info.lpVerb			= toStringz("open");
		info.lpFile			= toStringz("RUNDLL32.EXE");
		info.lpParameters	= toStringz("SHELL32.DLL,Control_RunDLL Inetcpl.cpl");
		info.lpDirectory	= null;
		info.nShow			= SW_SHOW;
/*
		char[1024] syspath;
		
		GetSystemDirectory(syspath.ptr, syspath.length);
		char[] rundll = std.path.join(std.windows.charset.fromMBSz(syspath.ptr), "Control.EXE");
		
		DaemonManager internet_option = new DaemonManager(rundll, "Inetcpl.cpl");
*/		
		scope(exit) {
//			EnableWindow(m_shell.handle, TRUE);
			m_shell.setEnabled(true);
			m_shell.setFocus();
		}
//		internet_option.start();
//		internet_option.setVisible(true);
		
//		HWND top = internet_option.getWindowHandle;
		
		ShellExecuteEx(&info);
		m_shell.setEnabled(false);
//		while (! internet_option.waitForSingleObject(0)) {
		while (WaitForSingleObject(info.hProcess, 0) == WAIT_TIMEOUT) {
			if (! m_shell.isDisposed()) {
				if (!m_display.readAndDispatch()) {
					m_display.sleep();
				}
			}
		}
	}
	
	void runDaemon()
	{
		try {
			m_daemon = new DaemonManager(getConfig.saku);
			
			m_daemon.start();
//			m_daemon.disableWindowStyle(DaemonManager.WindowStyle.WS_SYSMENU);
//			m_daemon.setClassEnabled("Button", false);
		} catch (Exception e) {
			MessageBox.showMessage(std.string.format(
				"%s \n\n"
				"朔の実行中にエラーが発生したため朔の制御オプションをOFFにします\n"
				"オプションを再設定してください", e.toString));
			m_daemon = null;
			getConfig.controlDaemon = false;
		}
	}
	
	void doOptionDialog(SelectionEvent e) 
	{
		OptionDialog dialog = new OptionDialog;
		char[] path = getConfig.saku();
		
		dialog.run();
		if (m_daemon && !getConfig.controlDaemon) {
			m_daemon.kill();
			m_daemon = null;
		}else if (!m_daemon && getConfig.controlDaemon) {
			runDaemon();
		} else if (std.string.icmp(path, getConfig.saku) != 0) {
			if (m_daemon) {
				m_daemon.kill();
				m_daemon = null;
			}
			runDaemon();
		}
	}
	
public:
	
	void showShell(bool sw)
	{
		if (sw) {
			m_shell.setVisible(true);
			m_shell.setMinimized(false);
			m_shell.setActive();
			m_shell.setFocus();
		} else {
			m_shell.setVisible(false);
		}
	}
	
	void shellIconifyHandler(ShellEvent e)
	{
		showShell(false);
	}
	
	void trayClickHandler(SelectionEvent e)
	{
		showShell(true);
	}
	
	void run()
	{
		m_display = Display.getDefault();
		m_shell = new Shell(m_display);
		
		m_shell.handleShellIconify(null, &shellIconifyHandler);
		
		m_browser = new SakuBrowser(m_shell);
		initializeIcons(m_display); // load icons
		int onoff = getConfig.controlDaemon;
		
		if (getConfig.controlDaemon) {
			runDaemon();
		}
		scope(exit) {
			if (m_daemon) {
				m_daemon.kill();
			}
			m_display.dispose();
		}
		
		m_shell.setText(getVersionString());
		m_shell.setSize(700, 500);
		m_shell.setLayout(new FillLayout);
		m_shell.setImage(getIcon("hoven"));
		
		m_shell.handleShellClose(null, &closeHandler);
		
		Menu menu_bar = new Menu(m_shell, DWT.BAR);
		m_shell.setMenuBar(menu_bar);
		
		MenuItem file_menu = new MenuItem(menu_bar, DWT.CASCADE);
		file_menu.setText("ファイル(&F)");
		Menu menu = new Menu(file_menu);
		file_menu.setMenu(menu);
		
		MenuItem exit_item = new MenuItem(menu, DWT.PUSH);
		exit_item.setText("終了(&X)");
		exit_item.handleSelection(null, &exitHandler);
		
		MenuItem view_menu = new MenuItem(menu_bar, DWT.CASCADE);
		view_menu.setText("表示(&V)");
		menu = new Menu(view_menu);
		view_menu.setMenu(menu);
		with (new MenuItem(menu, DWT.CHECK)) {
			setText("TkSaku");
			setSelection(true);
			handleSelection(null, &visibleTkSakuHandler);
		}
		
		MenuItem tool_menu = new MenuItem(menu_bar, DWT.CASCADE);
		tool_menu.setText("ツール(&T)");
		menu = new Menu(tool_menu);
		tool_menu.setMenu(menu);
		
		with (new MenuItem(menu, DWT.PUSH)) {
			setText("オプション(&O)");
			handleSelection(null, &doOptionDialog);
		}
		new MenuItem(menu, DWT.SEPARATOR);
		with (new MenuItem(menu, DWT.PUSH)) {
			setText("インターネットオプション(&I)");
			handleSelection(null, &doInternetOptionDialog);
		}
		
		MenuItem help_menu = new MenuItem(menu_bar, DWT.CASCADE);
		help_menu.setText("ヘルプ(&F)");
		menu = new Menu(help_menu);
		help_menu.setMenu(menu);
		
		MenuItem about_item = new MenuItem(menu, DWT.PUSH);
		about_item.setText(getVersionString ~ " について(&A)");
		about_item.handleSelection(null, &doAboutDialog);
		
		createTray();
		
		m_browser.createWidgets();
		
		deserialize();
		
		m_shell.open();
		m_shell.layout();
		
		try {
			while (! m_shell.isDisposed()) {
				if (!m_display.readAndDispatch()) {
					m_display.sleep();
				}
			}
		} catch (Object e) {
			MessageBox.showMessage(e.toString, getVersionString(), m_shell);
		}
		
		return 0;
	}
	
	void createTrayMenu()
	{
		m_tray_menu		= new Menu(m_shell);
		MenuItem item;
		
		with (item = new MenuItem(m_tray_menu, DWT.PUSH)) 
		{
			setText("表示");
			handleSelection(null, &traySelectionHandler);
			m_tray_menu.setDefaultItem(item);
			
			setData(new WrapperObject!(TrayItemId)(TrayItemId.VISIBLE));
		}
		cast(void)new MenuItem(m_tray_menu, DWT.SEPARATOR);
		
		with (new MenuItem(m_tray_menu, DWT.PUSH)) 
		{
			setText("終了");
			handleSelection(null, &traySelectionHandler);
			
			setData(new WrapperObject!(TrayItemId)(TrayItemId.EXIT));
		}
		m_tray_menu.setVisible(false);
	}
	
	void createTray()
	{
		Tray tray		= m_shell.getDisplay.getSystemTray();
		m_tray_item		= new TrayItem(tray, DWT.NONE);
		
		m_tray_item.handleSelection(null, &trayClickHandler);
		m_tray_item.handleEvent(null, DWT.MenuDetect, &trayMenuHandler);
		m_tray_item.setToolTipText(getVersionString() ~ "\n動作中");
		m_tray_item.setImage(getIcon("hoven"));
		
		createTrayMenu();
	}
	
	void trayMenuHandler(Event e)
	{
		Point pt = m_shell.getDisplay.getCursorLocation();
		m_tray_menu.setLocation(pt.x, pt.y);
		m_tray_menu.setVisible(true);
	}
	
	void traySelectionHandler(SelectionEvent e)
	{
		TrayItemId id = (cast(WrapperObject!(TrayItemId))(e.widget.getData)).getData;
		
		switch (id) 
		{
		case TrayItemId.VISIBLE:
			showShell(true);
			break;
		case TrayItemId.EXIT:
			m_shell.close();
			break;
		default:
			break;
		}
	}
}


// About dialog

class AboutDialog
{
protected:
	void closeHandler(SelectionEvent e)
	{
		Shell shell = cast(Shell)e.cData;
		shell.close();
	}
	
	void okHandler(SelectionEvent e)
	{
		Shell shell = cast(Shell)e.cData;
		shell.close();
	}
	
public:
	this()
	{
	}
	
	void run()
	{
		// initialize
		
		// open dialog
		Display	display	= Display.getCurrent();
		Shell shell 	= new Shell(display, DWT.CENTER|DWT.BORDER|DWT.TITLE|DWT.APPLICATION_MODAL);
		
		shell.setText(getVersionString);
		shell.setLayout(new GridLayout(1, true));
		
		with (new CLabel(shell, DWT.NONE)) {
			setImage(getIcon("hoven_banner"));
			setLayoutData(new GridData(GridData.BEGINNING));
		}
		with (new Text(shell, DWT.MULTI|DWT.READ_ONLY)) {
			setText(format(
`-= %s =-

Copyright (C) 2005 Hoven Project
(仮) http://d.hatena.ne.jp/ultraist/

-= Special Thanks =-

新月プロジェクト:
	- http://shingetsu.info/

FAMFAMFAM:
	- http://www.famfamfam.com/
`
				, getVersionString));
			setLayoutData(new GridData(GridData.FILL_BOTH));
		}
		Button button;
		with (button = new Button(shell, DWT.FLAT)) {
			setText("OK");
			setLayoutData(new GridData(GridData.FILL_HORIZONTAL));
			handleSelection(shell, &okHandler);
			shell.setDefaultButton(button);
		}
		
		shell.pack();
		shell.layout();
		
		shell.setLocation((display.getBounds.width - shell.getSize.x) / 2,
                    	   (display.getBounds.height - shell.getSize.y) / 2);
		shell.open();
		// event loop
		while (! shell.isDisposed()) {
			if (!display.readAndDispatch()) {
				display.sleep();
			}
		}
	}
}


class OptionDialog
{
protected:
	Shell	m_shell;
	Text	m_host;
	Text	m_port;
	Button	m_daemon;
	Text	m_path;
	Button	m_path_push;
	bool	m_modify_path = false;
	
	void closeHandler(SelectionEvent e)
	{
		Shell shell = cast(Shell)e.cData;
		shell.close();
	}
	
	void okHandler(SelectionEvent e)
	{
		Shell shell = cast(Shell)e.cData;
		Configuration config = getConfig;
		
		try {
			config.nodePort = std.conv.toUint(m_port.getText);
		} catch (Exception ex) {
			MessageBox.showMessage("ポート: " ~ ex.toString);
			e.doit = false;
			return;
		}
		config.nodeName = m_host.getText();
		config.controlDaemon = m_daemon.getSelection ? true:false;
		if (m_modify_path) {
			config.sakuPath = m_path.getText();
		}
		
		shell.close();
	}
	
	void checkDaemonControl(SelectionEvent e)
	{
		Button check = cast(Button)e.widget;
		
		if (check.getSelection) {
			m_path.setEnabled(true);
			m_path_push.setEnabled(true);
		} else {
			m_path.setEnabled(false);
			m_path_push.setEnabled(false);
		}
	}
	
	void doFileDialog(SelectionEvent e)
	{
		Shell shell = cast(Shell)e.cData;
		FileDialog dialog = new FileDialog(shell);
		char[][] ext;
		ext ~= "*.exe";
		
		dialog.setFilterExtensions(ext);
		char[] path = dialog.open();
		
		if (path.length) {
			m_path.setText(path);
			m_modify_path = true;
		}
	}
	
	void portInputHandler(KeyEvent e)
	{
		if (std.ctype.isalpha(e.character)
			|| std.ctype.isspace(e.character)) 
		{
			e.doit = false;
		}
		
	}
	
public:
	this(){}
	
	void run(Point point = null)
	{
		
		// open dialog
		Display	display	= Display.getCurrent();
		Shell shell 	= new Shell(display, DWT.CLOSE|DWT.BORDER|DWT.TITLE|DWT.APPLICATION_MODAL);
		
		shell.setText("オプション");
		shell.setLayout(new GridLayout(1, false));
		
		Group group;
		with (group = new Group(shell, DWT.NONE)) {
			setText("ゲートウェイ");
			setLayoutData(new GridData(GridData.FILL_BOTH));
			setLayout(new GridLayout(2, false));
		}
		with (new Label(group, DWT.NONE)) {
			setText("ホスト");
			setLayoutData(new GridData(GridData.BEGINNING));
		}
		with (m_host = new Text(group, DWT.BORDER|DWT.SINGLE)) {
			setText(getConfig.nodeName);
			setLayoutData(new GridData(200, -1));
		}
		with (new Label(group, DWT.NONE)) {
			setText("ポート");
			setLayoutData(new GridData(GridData.BEGINNING));
		}
		with (m_port = new Text(group, DWT.BORDER|DWT.SINGLE)) {
			setText(.std.string.toString(getConfig.nodePort));
			setLayoutData(new GridData(100, -1));
			handleKeyDown(null, &portInputHandler);
		}
		with (group = new Group(shell, DWT.NONE)) {
			setText("朔-WinSaku");
			setLayoutData(new GridData(GridData.FILL_BOTH));
			setLayout(new GridLayout(3, false));
		}
		
		with (m_daemon = new Button(group, DWT.CHECK)) {
			setText("朔を制御する");
			GridData grid = new GridData(GridData.FILL_HORIZONTAL);
			grid.horizontalSpan = 3;
			setLayoutData(grid);
			
			handleSelection(null, &checkDaemonControl);
			
			setSelection(getConfig.controlDaemon ? 1:0);
		}
		with (new Label(group, DWT.NONE)) {
			setText("朔");
			GridData grid = new GridData(GridData.BEGINNING);
			setLayoutData(grid);
		}
		with (m_path = new Text(group, DWT.BORDER|DWT.SINGLE|DWT.READ_ONLY)) {
			setText(getConfig.saku);
			setLayoutData(new GridData(300, -1));
			if (!m_daemon.getSelection) {
				setEnabled(false);
			}
		}
		with (m_path_push = new Button(group, DWT.FLAT)) {
			setText("...");
			if (!m_daemon.getSelection) {
				setEnabled(false);
			}
			handleSelection(shell, &doFileDialog);
		}
		
		Composite page;
		with (page = new Composite(shell, DWT.NONE)) {
			GridData grid = new GridData(GridData.HORIZONTAL_ALIGN_END);
			setLayoutData(grid);
			
			setLayout(new GridLayout(2, true));
		}
		Button set_button;
		with (set_button = new Button(page, DWT.PUSH)) {
			setText("設定");
			setLayoutData(new GridData(GridData.END));
			handleSelection(shell, &okHandler);
			// defualt button
			shell.setDefaultButton(set_button);
		}
		
		Button close_button;
		with (close_button = new Button(page, DWT.PUSH)) {
			setText("閉じる");
			setLayoutData(new GridData(GridData.END));
			
			handleSelection(shell, &closeHandler);
		}
		if (point !is null) {
			shell.setLocation(point);
		}
		shell.pack();
		shell.layout();
		shell.open();
		// event loop
		while (! shell.isDisposed()) {
			if (!display.readAndDispatch()) {
				display.sleep();
			}
		}
	}
}







