module pes.data.inifile;

private import std.windows.charset;
private import std.windows.syserror;
private import std.c.windows.windows;

version(Win32)
{

} else {
	pragma(msg, "warning: pes.data.inifile: not implemented for this compiler version");
}


extern(Windows)
{
	DWORD GetPrivateProfileStringA(
		LPCTSTR	lpAppName,
		LPCTSTR	lpKeyName,
		LPCTSTR	lpDefault,
		LPTSTR	lpReturnedString,
		DWORD	nSize,
		LPCTSTR	lpFileName
	);
	
	UINT GetPrivateProfileIntA(
		LPCTSTR	lpAppName,
		LPCTSTR	lpKeyName,
		INT		nDefault,
		LPCTSTR	lpFileName
	);
	
	BOOL WritePrivateProfileStringA(
		LPCTSTR	lpAppName,
		LPCTSTR	lpKeyName,
		LPCTSTR	lpString,
		LPCTSTR	lpFileName
	);
}
alias GetPrivateProfileStringA GetPrivateProfileString;
alias GetPrivateProfileIntA GetPrivateProfileInt;
alias WritePrivateProfileStringA WritePrivateProfileString;

class IniFileException: Exception
{
	this(char[] msg)
	{
		super(msg);
	}
}

class IniFile
{
protected:
	char[]	m_filepath;
	char[]	m_section;
	
	static void throwCurrentException()
	{
		DWORD ecd = GetLastError();
		if (ecd != 0) {
			throw new IniFileException(std.windows.syserror.sysErrorString(ecd));
		}
	}
	
public:
	static BUFFER_SIZE = 1024;
	
	this(char[] filepath)
	{
		m_filepath = filepath;
	}
	
	uint readInt(char[] section, char[] key)
	{
		return readInt(m_filepath, section, key);
	}
	
	char[] readString(char[] section, char[] key)
	{
		return readString(m_filepath, section, key);
	}
	
	void writeInt(char[] section, char[] key, uint data)
	{
		writeInt(m_filepath, section, key, data);
	}
	
	void writeString(char[] section, char[] key, char[] data)
	{
		writeString(m_filepath, section, key, data);
	}
	
	// 
	uint readInt(char[] key)
	{
		return readInt(m_filepath, m_section, key);
	}
	
	char[] readString(char[] key)
	{
		return readString(m_filepath, m_section, key);
	}
	
	void writeInt(char[] key, uint data)
	{
		writeInt(m_filepath, m_section, key, data);
	}
	
	void writeString(char[] key, char[] data)
	{
		writeString(m_filepath, m_section, key, data);
	}
	
	void setDefaultSection(char[] section)
	{
		m_section = section;
	}
	
	static void writeInt(char[] filepath, char[] section, char[] key, uint data)
	{
		writeString(filepath, section, key, std.string.toString(data));
	}
	
	static void writeString(char[] filepath, char[] section, char[] key, char[] data)
	{
		BOOL result;
		
		result = WritePrivateProfileString(
					std.windows.charset.toMBSz(section),
					std.windows.charset.toMBSz(key),
					std.windows.charset.toMBSz(data),
					std.windows.charset.toMBSz(filepath)
				);
		
		if (result == FALSE) {
			throwCurrentException();
		}
	}
	
	static char[] readString(char[] filepath, char[] section, char[] key)
	{
		DWORD result;
		uint buffer_size = BUFFER_SIZE;
		char* buffer = new char[buffer_size];
		
		result = GetPrivateProfileString(
					std.windows.charset.toMBSz(section),
					std.windows.charset.toMBSz(key),
					null,
					buffer,
					buffer_size,
					std.windows.charset.toMBSz(filepath)
				);
		
		if (result <= 0) {
			throwCurrentException();
		}
		
		return std.windows.charset.fromMBSz(buffer);
	}
	
	static uint readInt(char[] filepath, char[] section, char[] key)
	{
		DWORD result;
		
		result = GetPrivateProfileInt(
					std.windows.charset.toMBSz(section),
					std.windows.charset.toMBSz(key),
					-1,
					std.windows.charset.toMBSz(filepath)
				);
		
		if (result == -1) {
			throwCurrentException();
		}
		
		return cast(uint)result;
	}
	
}

debug(pes_data_inifile)
{
	import std.stdio;
	int main()
	{
		IniFile file = new IniFile("test.ini");
		file.setDefaultSection("test1");
		file.writeString("x", "漢字");
		file.writeInt("a", 0x0);
		file.writeInt("b", 0x8fffffff);
		file.writeInt("c", 0xffffffff);
		
		
		writefln("%s %s\n", file.readInt("c"), file.readString("x"));
		
		return 0;
	}
}


















