module pes.data.sqliteserializer;

/* SQLiteを使った変数保存機 */

private import std.string;
private import pes.data.sqlite3;
private import pes.variant;


//remark; 初めてのdeserializeではSQLExceptionを返すことがある
private import std.stdio;

class SQLiteSerializer
{
protected:
	Variant[uint]	m_member;
	char[]			m_db_name;
	char[]			m_table_name;
	
	char[] makeCreateSQL(long count)
	in {
		assert(count != 0);
	}
	body {
		char[] sql = "create table " ~ m_table_name ~ " (ID primary key,V0";
		for (long i = 1; i < count; ++i) {
			sql ~= ", V" ~ std.string.toString(i);
		}
		sql ~= ");";
		return sql;
	}
	
	char[] makeInsertSQL(VArray array)
	{
		char[] sql = "insert into " ~ m_table_name ~ " values('" ~ m_table_name ~ "',";
		long count = array.toLong;
		
		assert(count != 0);
		sql ~= "'" ~ escapeSQLText(array[0].toString) ~ "'";
		
		for (long i = 1; i < count; ++i) {
			sql ~= ", '" ~ escapeSQLText(array[i].toString) ~ "'";
		}
		sql ~= ")";
		
		return sql;
	}
	
	char[] makeUpdateSQL(VArray array)
	{
		char[] sql = "update " ~ m_table_name ~ " set ";
		long count = array.toLong;
		
		assert(count != 0);
		sql ~= "V0='" ~ escapeSQLText(array[0].toString) ~ "'";
		
		for (long i = 1; i < count; ++i) {
			sql ~= ", V" ~  std.string.toString(i) ~ "= '" ~ escapeSQLText(array[i].toString) ~ "'";
		}
		sql ~= " where ID= '" ~ m_table_name ~"'";
		
		return sql;
	}
	
	char[] makeDeleteSQL()
	{
		return "delete from " ~ m_table_name;
	}
	
public:
	this(char[] db_name, char[] table_name)
	{
		m_db_name 		= db_name;
		m_table_name	= table_name.dup;
	}
	
	void serialize(VArray array)
	{
		bool update = false;
		SQLite db = new SQLite(m_db_name);
		
		db.begin();
		scope(exit) {
			if (db !is null) {
				db.commit();
				db.close();
			}
		}
		char[] create_sql = makeCreateSQL(array.toLong);
		char[] insert_sql = makeInsertSQL(array);
		char[] update_sql = makeUpdateSQL(array);
//		char[] delete_sql = makeDeleteSQL();
		
//		writef("%s\n", insert_sql);
		
		try {
			// とりあえずアップデート
			db.exec(update_sql);
			update = true;
		} catch (SQLiteException) {
			// do nothing
			// 排他が必要な場合は、SQLiteBusyExceptionで動作変更しないとダメ
		}
		if (!update) {
			// アップデートできない場合は初めてと認定
			db.exec(create_sql);	// テーブル作成
			db.exec(insert_sql);	// インサート
		}
	}
	
	VArray deserialize()
	{
		char[] 			select_sql 	= "select * from " ~ m_table_name;
		SQLite 			db	 		= new SQLite(m_db_name);
		SQLiteStatement	stmt		= db.prepare(select_sql);
		Variant[] array;
		
		scope(exit) {
			if (stmt !is null) {
				stmt.finalize();
			}
			if (db !is null) {
				db.close();
			}
		}
		stmt.step();
		for (int i = 1; i < stmt.columnCount(); ++i) { // 0はID
			array ~= Variant.make(stmt.fetch!(char[])(i).dup);
		}
		
		return new VArray(array);
	}
}


