module pes.net.http.header;

private import std.string;
private import pes.net.http.status;

public const char[2] CRLF = [ 0x0d, 0x0a ];
public const uint CRLF_LENGTH = 2; // CRLF.length

class HTTPHeaderField
{
protected:
	char[]	m_name;
	char[]	m_value;
	
public:
	const char[2] SEPARATOR = ": ";
	const uint SEPARATOR_LENGTH = 2;
	
	this(){}
	this(char[] name, char[] value)
	{
		m_name = name;
		m_value = value;
	}
	
	void setName(char[] name)
	{
		m_name = name;
	}
	
	void setValue(char[] value)
	{
		m_value = value;
	}
	
	char[] getName()
	{
		return m_name;
	}
	
	char[] getValue()
	{
		return m_value;
	}
	
	override char[] toString()
	{
		char[] field = new char[computeLength];
		uint pos1 = m_name.length;
		uint pos2 = pos1 + SEPARATOR_LENGTH;
		uint pos3 = pos2 + m_value.length;
		
		field[0 .. pos1] 	= m_name;
		field[pos1 .. pos2]	= SEPARATOR;
		field[pos2 .. pos3]	= m_value;
		
		return field;
	}
	
	uint computeLength()
	{
		return m_name.length + SEPARATOR_LENGTH + m_value.length;
	}
	
	void clear()
	{
		m_name = null;
		m_value = null;
	}
}

class HTTPHeader
{
protected:
	HTTPStatusCode				m_status;
	float						m_version;
	HTTPHeaderField[char[]]		m_fields;
	
public:
	this() {}
	
	this(float http_version)
	{
		m_version = http_version;
	}
	
	this(HTTPHeaderField[] fields)
	{
		setFields(fields);
	}
	
	void setFields(HTTPHeaderField[] fields)
	{
		m_fields = null;
		foreach (field; fields) {
			addField(field);
		}
	}
	
	void addField(HTTPHeaderField field)
	{
		m_fields[std.string.toupper(field.getName)] = field;
	}
	
	void addField(char[] name, char[] value)
	{
		addField(new HTTPHeaderField(name, value));
	}
	
	HTTPHeaderField getField(char[] name)
	{
		HTTPHeaderField *field =  (std.string.toupper(name) in m_fields);
		if (field == null) {
			return null;
		}
		
		return *field;
	}
	
	char[] getFieldValue(char[] name)
	{
		HTTPHeaderField field = getField(name);
		
		if (field is null) {
			return null;
		}
		
		return field.getValue;
	}
	
	HTTPHeaderField[] getFields()
	{
		return m_fields.values;
	}
	
	void clear()
	{
		foreach (inout field; m_fields) {
			field.clear();
			field = null;
		}
		m_fields = null;
	}
	
	override char[] toString()
	{
		char[] field_str;
		char[] header = new char[computeLength];
		uint pos1 = 0;
		uint pos2 = 0;
		
		foreach(field; getFields()) {
			pos2 = pos1 + field.computeLength;
			field_str = field.toString;
			header[pos1 .. pos2] = field_str;
			delete field_str;
			pos1 = pos2;
			pos2 += CRLF_LENGTH;
			header[pos1 .. pos2] = CRLF;
			pos1 = pos2;
		}
		header[pos1 .. pos1 + CRLF_LENGTH] = CRLF;
		field_str = null;
		
		return header;
	}
	
	uint computeLength()
	{
		uint header_length 	= 0;
		
		foreach(field; getFields()) {
			header_length += field.computeLength + CRLF_LENGTH;
		}
		
		return header_length + CRLF_LENGTH;
	}
	
	void setStatusCode(HTTPStatusCode status)
	{
		m_status = status;
	}
	
	HTTPStatusCode getStatusCode()
	{
		return m_status;
	}
	
	void setVersion(float http_version)
	{
		m_version = http_version;
	}
	
	float getVersion()
	{
		return m_version;
	}
}


