module pes.net.http.responselistener;

private import pes.net.http.header;
private import pes.net.http.status;

class HTTPResponseEvent
{
protected:
	bool		m_disconnect;
	void[]		m_data;
	HTTPHeader	m_header;
	
public:
	void header(HTTPHeader header)
	{
		m_header = header;
	}
	
	HTTPHeader header()
	{
		return m_header;
	}
	
	void disconnect(bool sw)
	{
		m_disconnect = sw;
	}
	
	bool disconnect()
	{
		return m_disconnect;
	}
	
	void data(void[] data)
	{
		m_data = data;
	}
	
	char[] text()
	{
		return cast(char[])m_data;
	}
	
	char[] duplicationText()
	{
		return cast(char[])m_data.dup;
	}
	
	void[] duplicationData()
	{
		return data.dup;
	}
	
	void[] data()
	{
		return m_data;
	}
	
	void reset()
	{
		m_disconnect	= false;
		m_data			= null;
		m_header		= null;
	}
}

class HTTPResponseException: Exception
{
protected:
	HTTPStatusCode m_status;
	
public:
	this(char[]msg, HTTPStatusCode status = HTTPStatusCode.UNDEF)
	{
		super(std.string.format(msg));
		m_status = status;
	}
	
	HTTPStatusCode getStatusCode()
	{
		return m_status;
	}
}

class HTTPLineOverflowException: HTTPResponseException
{
	this()
	{
		super(this.classinfo.name);
	}
}

abstract class HTTPResponseListener
{
protected:
	uint	m_bucket_size		= 0x2000;	// 8KB
	bool	m_line_enabled		= false;
	
public:
	void begin(HTTPResponseEvent header_event) {}
	void end() {}
	
	abstract void receiving(HTTPResponseEvent event);
	void receivingLine(HTTPResponseEvent event) {}
	
	void setLineHandlerEnabled(bool sw)
	{
		m_line_enabled = sw;
	}
	
	bool getLineHandlerEnabled()
	{
		return m_line_enabled;
	}
	
	void setBucketSize(uint bucket_size)
	{
		m_bucket_size = bucket_size;
	}
	
	uint getBucketSize()
	{
		return m_bucket_size;
	}
}

