module pes.net.shingetsu.gatewayio;

private import std.string;
private import std.conv;
private import std.socket;
private import std.c.time;
private import pes.net.http.all;
private import pes.net.shingetsu.get;

class GatewayException: Exception
{
	this(char[] msg) {
		super(msg);
	}
}


class ThreadResponseEvent
{
public:
	char[]	name;
	time_t	lastModifiedTime;
	char[]	path;
	char[]	uri;
	char[]	type;
	char[]	title;
	uint	records;
	uint	threadSize;
	
	Object	cData;
	bool	disconnect;
	
	this(){ }
	
	this(char[] csv, Object data)
	{
		assign(csv, data);
	}
	
	this(char[][] parts, Object data)
	{
		assign(parts, data);
	}
	
	
	void assign(char[] csv, Object data)
	{
		assign(std.string.split(csv, ","), data);
	}
	
	void assign(char[][] parts, Object data)
	in {
		debug {
			assert(parts.length == 8);
		}
	}
	body {
		// format check
		if (parts.length != 8) {
			throw new GatewayException("format error");
		}
		std.utf.validate(parts[5]);
		
		// set
		name				= parts[0];
		lastModifiedTime 	= cast(time_t)std.conv.toUint(parts[1]);
		path				= parts[2];
		uri					= parts[3];
		type				= parts[4];
		title				= parts[5];
		records				= std.conv.toUint(parts[6]);
		threadSize			= std.conv.toUint(std.string.chomp(parts[7]));
		
		disconnect			= false;
		cData				= data;
	}
	
	void clear()
	{
		name				= null;
		lastModifiedTime 	= 0;
		path				= null;
		uri					= null;
		type				= null;
		title				= null;
		records				= 0;
		threadSize			= 0;
		
		cData				= null;
		disconnect			= false;
	}
}

class GatewayIO: ShinGETsuIO
{
protected:
	class ThreadListRequest: SingleGetRequest
	{
		this(char[] type)
		{
			super(std.string.format(
				"/gateway.cgi/csv/%s/file,stamp,path,uri,type,title,records,size",
				type)
			);
		}
	}
	
	class ThreadListResponseListener: HTTPResponseListener
	{
	protected:
		ThreadResponseEvent 				m_event;
		void delegate(ThreadResponseEvent)	m_handler;
		Object								m_argv;
	public:
		
		this(Object argv,  void delegate(ThreadResponseEvent) handler)
		{
			m_handler 	= handler;
			m_argv		= argv;
			setLineHandlerEnabled(true);
			
			m_event = new ThreadResponseEvent;
			m_event.clear();
		}
		
		override void receivingLine(HTTPResponseEvent event)
		{
			try {
				m_event.assign(event.text, m_argv);
			} catch (Exception e) {
				m_event.clear();
				return;
			}
			m_handler(m_event);
			event.disconnect = m_event.disconnect;
			
			m_event.clear();
		}
		
		override void receiving(HTTPResponseEvent event)
		{
			// do nothing
		}
	}
	
public:
	this(InternetAddress address)
	{
		super(address);
	}
	
	void getRegisteredThreads(Object argv, void delegate(ThreadResponseEvent) handler)
	{
		ThreadListRequest 			generator 	= new ThreadListRequest("changes");
		ThreadListResponseListener 	listener	= new ThreadListResponseListener(argv, handler);
		
		scope(exit) {
			removeListener();
		}
		setListener(listener);
		
		get(generator);
		
		delete listener;
		delete generator;
	}
	
	void getRecentThreads(Object argv, void delegate(ThreadResponseEvent) handler)
	{
		ThreadListRequest 			generator 	= new ThreadListRequest("recent");
		ThreadListResponseListener 	listener	= new ThreadListResponseListener(argv, handler);
		
		scope(exit) {
			removeListener();
		}
		setListener(listener);
		
		get(generator);
		
		delete listener;
		delete generator;
	}
}








