/**
 * スキーマ	
 *
 * shingetsu://URLの扱い
 */
module schema;

private import std.string;
private import std.conv;
private import pes.net.shingetsu.util;

private import config;

class SchemaException: Exception
{
	this(char[] msg)
	{
		super(msg);
	}
}

class Schema
{
	enum Mode :uint{
		UNDEF		= 0,
		BACK_ROWS	= 1,
		PAGES		= 2,
	};
	
	uint	page;
	uint	items;
	Mode	mode;
	char[]	threadName;
	char[]	method;
	char[]	hash;
	char[]	unknown;
	char[]	top;
	
	this() {}
	
	this(char[] title)
	{
		threadName = title;
	}
	
	char[] toURI()
	{
		char[] param;
		char[] top;
		
		switch (mode) {
		case Mode.BACK_ROWS:
			param = std.string.format("?mode=%s&items=%s", cast(uint)mode, items);
			break;
		case Mode.PAGES:
			param = std.string.format("?mode=%s&items=%s&page=%s", cast(uint)mode, items, page);
			break;
		default:
			break;
		}
		if (top.length != 0) {
			top = std.string.format("#%s", top);
		}
		
		return std.string.format("shingetsu://%s/%s/%s%s%s", 
								encodeThreadName(threadName), method, hash, param, top);
	}
	
	static Schema parse(char[] uri)
	in {
		assert(std.string.ifind(uri, "shingetsu://") == 0);
	}
	body {
		Schema schema = new Schema;
		char[][] parts;
		char[] param;
		
		if (std.string.ifind(uri, "shingetsu://") != 0) {
			throw new SchemaException("parse error");
		}
		
		parts = std.string.split(uri, "?");
		if (parts.length > 1) {
			uri				= parts[0];
			param			= parts[1];
			
			parts = std.string.split(param, "#");
			if (parts.length > 1) {
				param			= parts[0];
				schema.top		= parts[1];
			}
		} else {
			parts = std.string.split(uri, "#");
			if (parts.length > 1) {
				uri				= parts[0];
				schema.top		= parts[1];
			}
		}
		
		parts = std.string.split(uri, "/");
		if (parts.length < 4) {
			throw new SchemaException("parse error");
		}
		schema.threadName 	= decodeThreadName(parts[2]);
		schema.method		= parts[3];
		
		if (parts.length > 4) {
			if (parts[4].length != 0) {
				schema.hash = parts[4];
			}
		}
		schema.page 		= 0;				// default
		schema.items		= getConfig.items;	// default
		schema.mode			= Mode.UNDEF;		// default
		
		if (param.length != 0) {
			char[][] pairs = std.string.split(param, "&");
			
			foreach (pair; pairs) {
				char[][] kv = std.string.split(pair, "=");
				if (kv.length == 2) {
					if (std.string.cmp(kv[0], "page") == 0) {
						try {
							schema.page = std.conv.toUint(kv[1]);
						} catch (ConvError) {
							// do nothing
						}
					} else if (std.string.cmp(kv[0], "items") == 0) {
						try {
							schema.items = std.conv.toUint(kv[1]);
						} catch (ConvError) {
							// do nothing
						}
					} else if (std.string.cmp(kv[0], "mode") == 0) {
						try {
							uint mode = std.conv.toUint(kv[1]);
							switch (mode) {
							case Mode.BACK_ROWS:
								schema.mode = Mode.BACK_ROWS;
								break;
							case Mode.PAGES:
								schema.mode = Mode.PAGES;
								break;
							default:
								schema.mode = Mode.UNDEF;
								break;
							}
						} catch (ConvError) {
							// do nothing
						}
					}
				}
			}
		}
		
		return schema;
	}
}
